import logging
from copy import copy
from openpyxl import load_workbook
from datetime import date
import pandas as pd
from pandas import DataFrame

from mario.data_extractor import DataExtractor
from mario.dataset_specification import DatasetSpecification
from mario.metadata import Metadata

logger = logging.getLogger(__name__)
style_attrs = ["alignment", "border", "fill", "font", "number_format", "protection"]


def rotate_list(list_1, n):
    """
    Rotates a list n positions. Used to push measure to the front of the data
    :param list_1: the list to rotate
    :param n: number of positions to rotate
    :return: rotated list
    """
    return list_1[-n:] + list_1[:-n]


class ExcelBuilder(object):
    """
    Class for creating Excel outputs
    """

    def __init__(self,
                 output_file_path: str,
                 data_extractor: DataExtractor,
                 dataset_specification: DatasetSpecification,
                 metadata: Metadata,
                 template_path: str
                 ):
        self.data_extractor = data_extractor
        self.dataset_specification = dataset_specification
        self.metadata = metadata
        self.filepath = output_file_path
        self.template = "excel_template.xlsx"
        if template_path is not None:
            self.template = template_path
        self.workbook = None
        self.rows = None
        self.cols = None
        self.total = 0
        logger.debug("Excel handler initialised")

    def create_workbook(self, create_notes_page=False):
        """
        Creates a write-only workbook and builds
        content in streaming mode to conserve memory
        """
        template_workbook = load_workbook(self.template)
        template_workbook.save(self.filepath)

        if create_notes_page:
            self.__create_notes_page__()
        else:
            self.__delete_notes_page__()
        self.__create_data_page__()
        self.__create_pivot_page__()

        # Set active sheet
        workbook = load_workbook(self.filepath)
        for sheet in workbook:
            workbook[sheet.title].views.sheetView[0].tabSelected = False
        workbook.save(self.filepath)
        workbook.close()

        logger.debug("Completed workbook")

    def __format_cell__(self, cell, new_cell):
        """
        Copy attributes from src to dst. Attributes are shallow-copied to avoid
        TypeError: unhashable type: 'StyleProxy'
        """
        for name in style_attrs:
            setattr(new_cell, name, copy(getattr(cell, name)))
        return new_cell

    def __replace_template_values(self, cell):
        """
        Replaces template values in a cell with metadata from the order
        """
        value = cell.value
        if value == '{ENQUIRY_NUMBER}':
            value = self.dataset_specification.collection
        if value == '{CUSTOMER_REFERENCE}':
            value = self.dataset_specification.get_property('customerRef')
        if value == '{USER}':
            value = 'Jisc Tailored Datasets App v1.0'
        if value == '{DATE}':
            value = date.today().isoformat()
        if value == '{DATASOURCE}':
            value = self.dataset_specification.get_property('datasource')
        if value == '{POP}':
            value = self.total
        if value == '{CAT}':
            value = self.dataset_specification.get_property('onwardUseCategory')
        if value == '{FORMAT}':
            value = 'Excel pivot table'
        return value

    def __create_notes_page__(self):
        workbook = load_workbook(self.filepath)
        notes = workbook.get_sheet_by_name('Notes')
        self.__update_notes__(notes)
        workbook.save(self.filepath)

    def __delete_notes_page__(self):
        workbook = load_workbook(self.filepath)
        if 'Notes' in workbook.sheetnames:
            workbook.remove(workbook['Notes'])
            workbook.save(self.filepath)

    def __create_pivot_page__(self):
        workbook = load_workbook(self.filepath)
        pivot = workbook['Pivot']._pivots[0]
        # Update the pivot table range
        pivot.cache.cacheSource.worksheetSource.ref = self.__range__()
        # Set to refresh
        pivot.cache.refreshOnLoad = True
        workbook.save(self.filepath)

    def create_notes_only(self, filename=None, data_format='CSV'):
        """
        Create a Notes page only, to accompany CSV outputs
        """
        self.workbook = load_workbook(self.template)
        self.__update_notes__(data_format=data_format, ws=self.workbook.get_sheet_by_name('Notes'))
        self.workbook.remove(self.workbook.get_sheet_by_name('Data'))
        self.workbook.remove(self.workbook.get_sheet_by_name('Pivot'))
        if filename is not None:
            self.workbook.save(filename=filename)
        else:
            self.workbook.save(filename=self.filepath)

    def __create_data_page__(self):
        df: DataFrame = self.data_extractor.get_data_frame()

        # Reorder the columns to put the measure in col #1
        cols = df.columns.tolist()
        cols = rotate_list(cols, 1)
        df = df[cols]

        self.rows = len(df)
        self.cols = len(cols)

        with pd.ExcelWriter(
                self.filepath,
                engine='openpyxl',
                mode="a",
                if_sheet_exists='replace'
        ) as writer:
            df.to_excel(excel_writer=writer, sheet_name='Data', index=False)

    def __update_notes__(self, ws, data_format='Excel pivot table'):
        """
        Updates the notes page
        """
        self.total = self.data_extractor.get_total()

        ws['B6'] = self.dataset_specification.collection
        ws['B7'] = self.dataset_specification.get_property('customerRef')
        ws['B9'] = 'Jisc Tailored Datasets App v1.0'
        ws['B11'] = date.today().isoformat()
        ws['B14'] = self.dataset_specification.get_property('datasource')
        ws['B15'] = self.total
        ws['B16'] = self.dataset_specification.get_property('onwardUseCategory')
        ws['B17'] = data_format

        # Field names and definitions
        row = 32
        for item in self.dataset_specification.items:
            meta = self.metadata.get_metadata(item)
            ws['A' + str(row)] = item
            if meta.get_property('tdsaDescription'):
                ws['B' + str(row)] = meta.get_property('tdsaDescription')
            row += 1

    def __range__(self):
        """
        Determine the cell range based on the dataset size
        :return: a cell range e.g "A1:C26"
        """
        return 'A1' + ':' + chr(ord('@') + self.cols) + str(self.rows)



import os

import pandas as pd
import pytest

from mario.data_extractor import Configuration, HyperFile, DataExtractor, StreamingDataExtractor
from mario.dataset_builder import DatasetBuilder, Format
from mario.dataset_specification import dataset_from_json, Constraint, dataset_from_manifest
from mario.metadata import metadata_from_json, metadata_from_manifest
from mario.query_builder import SubsetQueryBuilder, ViewBasedQueryBuilder


def test_integration_tdsx():
    dataset = dataset_from_json(os.path.join('test', 'dataset.json'))
    metadata = metadata_from_json(os.path.join('test', 'metadata.json'))
    configuration = Configuration(file_path=os.path.join('test', 'orders.hyper'))
    extractor = HyperFile(configuration=configuration, dataset_specification=dataset, metadata=metadata)
    builder = DatasetBuilder(dataset_specification=dataset, metadata=metadata, data=extractor)
    path = os.path.join('output', dataset.collection, dataset.name + '.tdsx')
    os.makedirs(os.path.join('output', dataset.collection), exist_ok=True)
    builder.build(file_path=path, output_format=Format.TABLEAU_PACKAGED_DATASOURCE)


def test_integration_csv():
    dataset = dataset_from_json(os.path.join('test', 'dataset.json'))
    metadata = metadata_from_json(os.path.join('test', 'metadata.json'))
    configuration = Configuration(file_path=os.path.join('test', 'orders.hyper'))
    extractor = HyperFile(configuration=configuration, dataset_specification=dataset, metadata=metadata)
    builder = DatasetBuilder(dataset_specification=dataset, metadata=metadata, data=extractor)
    path = os.path.join('output', dataset.collection, dataset.name + '.csv')
    os.makedirs(os.path.join('output', dataset.collection), exist_ok=True)
    builder.build(file_path=path, output_format=Format.CSV)


def test_integration_excel():
    dataset = dataset_from_manifest(os.path.join('test', 'manifest_superstore.json'))
    metadata = metadata_from_manifest(os.path.join('test', 'manifest_superstore.json'))
    configuration = Configuration(file_path=os.path.join('test', 'orders.hyper'))
    extractor = DataExtractor(configuration=configuration, dataset_specification=dataset, metadata=metadata)
    builder = DatasetBuilder(dataset_specification=dataset, metadata=metadata, data=extractor)
    path = os.path.join('output', 'test_integration_excel', 'data.xlsx')
    os.makedirs(os.path.join('output', 'test_integration_excel'), exist_ok=True)
    builder.build(file_path=path, output_format=Format.EXCEL_PIVOT, template_path='excel_template.xlsx')


def test_integration_excel_no_notes():
    from openpyxl import load_workbook
    dataset = dataset_from_manifest(os.path.join('test', 'manifest_superstore.json'))
    metadata = metadata_from_manifest(os.path.join('test', 'manifest_superstore.json'))
    configuration = Configuration(file_path=os.path.join('test', 'orders.hyper'))
    extractor = DataExtractor(configuration=configuration, dataset_specification=dataset, metadata=metadata)
    builder = DatasetBuilder(dataset_specification=dataset, metadata=metadata, data=extractor)
    path = os.path.join('output', 'test_integration_excel_no_notes', 'data.xlsx')
    os.makedirs(os.path.join('output', 'test_integration_excel_no_notes'), exist_ok=True)
    builder.build(file_path=path, output_format=Format.EXCEL_PIVOT, template_path='excel_template.xlsx')

    workbook = load_workbook(filename=path)
    assert 'Notes' not in workbook.sheetnames


def test_integration_excel_info_only():
    dataset = dataset_from_manifest(os.path.join('test', 'manifest_superstore.json'))
    metadata = metadata_from_manifest(os.path.join('test', 'manifest_superstore.json'))
    configuration = Configuration(file_path=os.path.join('test', 'orders.hyper'))
    extractor = DataExtractor(configuration=configuration, dataset_specification=dataset, metadata=metadata)
    builder = DatasetBuilder(dataset_specification=dataset, metadata=metadata, data=extractor)
    path = os.path.join('output', 'test_integration_excel_info', 'info.xlsx')
    os.makedirs(os.path.join('output', 'test_integration_excel_info'), exist_ok=True)
    builder.build(file_path=path, output_format=Format.EXCEL_INFO_SHEET, template_path='excel_template.xlsx')


def test_sql_extraction():
    # Skip this test if we don't have a connection string
    if not os.environ.get('CONNECTION_STRING'):
        pytest.skip("Skipping SQL test as no database configured")

    # Set up local test database, drivers and connection string to run this
    dataset = dataset_from_json(os.path.join('test', 'dataset.json'))
    metadata = metadata_from_json(os.path.join('test', 'metadata.json'))
    # Add constraints so we test parameter generation
    constraint = Constraint()
    constraint.item = 'Ship Mode'
    constraint.allowed_values = ['First Class']
    dataset.constraints.append(constraint)
    constraint = Constraint()
    constraint.item = 'City'
    constraint.allowed_values = ['Dallas']
    dataset.constraints.append(constraint)

    configuration = Configuration(
        connection_string=os.environ.get("CONNECTION_STRING"),
        schema='dev',
        view='superstore',
        query_builder=SubsetQueryBuilder
    )
    extractor = DataExtractor(
        dataset_specification=dataset,
        metadata=metadata,
        configuration=configuration
    )
    extractor.validate_data()
    sql_path = os.path.join('output', 'test_sql_extraction', 'query.sql')
    csv_path = os.path.join('output', 'test_sql_extraction', 'data.csv')
    os.makedirs(os.path.join('output', 'test_sql_extraction'), exist_ok=True)
    extractor.save_data_as_csv(csv_path)
    extractor.save_query(sql_path)


def test_sql_extraction_using_manifest():
    # Skip this test if we don't have a connection string
    if not os.environ.get('CONNECTION_STRING'):
        pytest.skip("Skipping SQL test as no database configured")

    # Set up local test database, drivers and connection string to run this
    dataset = dataset_from_manifest(os.path.join('test', 'manifest_superstore.json'))
    metadata = metadata_from_manifest(os.path.join('test', 'manifest_superstore.json'))
    configuration = Configuration(
        connection_string=os.environ.get("CONNECTION_STRING"),
        schema='dev',
        view='superstore',
        query_builder=SubsetQueryBuilder
    )
    extractor = DataExtractor(
        dataset_specification=dataset,
        metadata=metadata,
        configuration=configuration
    )
    extractor.validate_data()
    sql_path = os.path.join('output', 'test_sql_extraction_using_manifest', 'query.sql')
    csv_path = os.path.join('output', 'test_sql_extraction_using_manifest', 'data.csv')
    os.makedirs(os.path.join('output', 'test_sql_extraction_using_manifest'), exist_ok=True)
    extractor.save_data_as_csv(csv_path)
    extractor.save_query(sql_path)


def test_integration_excel_info_only_with_totals():
    from openpyxl import Workbook, load_workbook
    dataset = dataset_from_manifest(os.path.join('test', 'manifest_superstore.json'))
    metadata = metadata_from_manifest(os.path.join('test', 'manifest_superstore.json'))
    configuration = Configuration(file_path=os.path.join('test', 'orders.hyper'))
    extractor = DataExtractor(configuration=configuration, dataset_specification=dataset, metadata=metadata)
    total_from_query = extractor.get_total()
    builder = DatasetBuilder(dataset_specification=dataset, metadata=metadata, data=extractor)
    path = os.path.join('output', 'test_integration_excel_info', 'info.xlsx')
    os.makedirs(os.path.join('output', 'test_integration_excel_info'), exist_ok=True)
    builder.build(file_path=path, output_format=Format.EXCEL_INFO_SHEET, template_path='excel_template.xlsx')
    workbook: Workbook = load_workbook(path)
    total_from_notes = workbook.get_sheet_by_name('Notes')['B15'].value
    assert total_from_notes == total_from_query


def test_integration_csv_streaming():
    # Skip this test if we don't have a connection string
    if not os.environ.get('CONNECTION_STRING'):
        pytest.skip("Skipping SQL test as no database configured")
    dataset = dataset_from_json(os.path.join('test', 'dataset.json'))
    dataset.collection = 'test_integration_csv_streaming'
    metadata = metadata_from_json(os.path.join('test', 'metadata.json'))
    configuration = Configuration(
        connection_string=os.environ.get('CONNECTION_STRING'),
        schema="dev",
        view="superstore",
        query_builder=ViewBasedQueryBuilder
    )
    extractor = StreamingDataExtractor(
        dataset_specification=dataset,
        metadata=metadata,
        configuration=configuration
    )
    builder = DatasetBuilder(dataset_specification=dataset, metadata=metadata, data=extractor)
    path = os.path.join('output', dataset.collection, dataset.name + '.csv')
    os.makedirs(os.path.join('output', dataset.collection), exist_ok=True)
    builder.build(file_path=path, output_format=Format.CSV)
    df = pd.read_csv(path)
    assert len(df) == 10194


def test_integration_tdsx_streaming():
    # Skip this test if we don't have a connection string
    if not os.environ.get('CONNECTION_STRING'):
        pytest.skip("Skipping SQL test as no database configured")
    dataset = dataset_from_json(os.path.join('test', 'dataset.json'))
    dataset.collection = 'test_integration_tdsx_streaming'
    metadata = metadata_from_json(os.path.join('test', 'metadata.json'))
    configuration = Configuration(
        connection_string=os.environ.get('CONNECTION_STRING'),
        schema="dev",
        view="superstore",
        query_builder=ViewBasedQueryBuilder
    )
    extractor = StreamingDataExtractor(
        dataset_specification=dataset,
        metadata=metadata,
        configuration=configuration
    )
    builder = DatasetBuilder(dataset_specification=dataset, metadata=metadata, data=extractor)
    path = os.path.join('output', dataset.collection, dataset.name + '.tdsx')
    os.makedirs(os.path.join('output', dataset.collection), exist_ok=True)
    builder.build(file_path=path, output_format=Format.TABLEAU_PACKAGED_DATASOURCE)


def test_remove_redundant_hierarchies():
    metadata_file = os.path.join('test', 'metadata.json')
    metadata = metadata_from_json(file_path=metadata_file)
    dataset = dataset_from_json(os.path.join('test', 'dataset.json'))
    dataset.collection = 'test_remove_redundant_hierarchies'

    # remove an item so we have a one-item hierarchy for "Product"
    dataset.dimensions.remove("Product Name")
    assert "Product Name" not in dataset.items

    configuration = Configuration(file_path=os.path.join('test', 'orders.hyper'))
    extractor = HyperFile(configuration=configuration, dataset_specification=dataset, metadata=metadata)
    builder = DatasetBuilder(dataset_specification=dataset, metadata=metadata, data=extractor)

    # remove the product hierarchy
    builder.remove_redundant_hierarchies()

    assert "Product" not in metadata.get_hierarchies()
    assert "Location" in metadata.get_hierarchies()

    path = os.path.join('output', dataset.collection, dataset.name + '.tdsx')
    os.makedirs(os.path.join('output', dataset.collection), exist_ok=True)
    builder.build(file_path=path, output_format=Format.TABLEAU_PACKAGED_DATASOURCE)


def test_hyper_with_csv_output():
    folder = os.path.join('output', 'test_integration_hyper_with_csv_output')
    os.makedirs(folder, exist_ok=True)
    dataset = dataset_from_json(os.path.join('test', 'dataset.json'))
    metadata = metadata_from_json(os.path.join('test', 'metadata.json'))
    configuration = Configuration(
        file_path=os.path.join('test', 'orders.hyper')
    )
    extractor = HyperFile(
        dataset_specification=dataset,
        metadata=metadata,
        configuration=configuration
    )
    builder = DatasetBuilder(
        dataset_specification=dataset,
        metadata=metadata,
        data=extractor
    )
    builder.build(Format.CSV, file_path=os.path.join(folder, 'test.csv'))
    builder.build(Format.CSV, file_path=os.path.join(folder, 'test.csv.gz'))
    builder.build(Format.TABLEAU_PACKAGED_DATASOURCE, file_path=os.path.join(folder, 'test.tdsx'))
    builder.build(Format.EXCEL_PIVOT, file_path=os.path.join(folder, 'test.xlsx'))


def test_hyper_with_compressed_csv_output():
    folder = os.path.join('output', 'test_integration_hyper_with_csv_output')
    os.makedirs(folder, exist_ok=True)
    dataset = dataset_from_json(os.path.join('test', 'dataset.json'))
    metadata = metadata_from_json(os.path.join('test', 'metadata.json'))
    configuration = Configuration(
        file_path=os.path.join('test', 'orders.hyper')
    )
    extractor = HyperFile(
        dataset_specification=dataset,
        metadata=metadata,
        configuration=configuration
    )
    builder = DatasetBuilder(
        dataset_specification=dataset,
        metadata=metadata,
        data=extractor
    )
    builder.build(Format.CSV, file_path=os.path.join(folder, 'test.csv'))
    builder.data.save_data_as_csv(file_path=os.path.join(folder, 'test.csv'), compress_using_gzip=True)
    assert os.path.exists(os.path.join(folder, 'test.csv.gz'))



import os

from mario.data_extractor import DataExtractor, Configuration
from mario.dataset_specification import dataset_from_json
from mario.dataset_builder import DatasetBuilder, Format
from mario.metadata import metadata_from_json

# Build a TD extract in Excel
dataset = dataset_from_json("data_spec.xlsx")
metadata = metadata_from_json('metadata.json')
configuration = Configuration(connection_string=os.environ.get('odbc_connection_string'))
extractor = DataExtractor(configuration=configuration, dataset_specification=dataset, metadata=metadata)
# Either put automation code in here, or override the base class
builder = DatasetBuilder(dataset_specification=dataset, metadata=metadata, data=extractor)
path = os.path.join(dataset.enquiry, dataset.name + '.xlsx')
builder.build(file_path=path, output_format=Format.EXCEL_PIVOT)

# Build a Heidi Plus Datasource in TDSX
dataset = dataset_from_json("dataset.json")
metadata = metadata_from_json('metadata.json')
configuration = Configuration(
    connection_string=os.environ.get('odbc_connection_string'),
    view='dbo.v_heidi_plus_student_fpe'
)
extractor = DataExtractor(configuration=configuration, dataset_specification=dataset, metadata=metadata)
builder = DatasetBuilder(dataset_specification=dataset, metadata=metadata, data=extractor)
path = os.path.join('datasources', dataset.collection, dataset.name + '.tdsx')
builder.build(file_path=path, output_format=Format.TABLEAU_PACKAGED_DATASOURCE)



import os
import shutil
import tempfile
import logging

import pandas as pd
import pytest

from mario.data_extractor import DataExtractor, Configuration, StreamingDataExtractor, DataFrameExtractor, \
    PartitioningExtractor, HyperFile
from mario.dataset_specification import dataset_from_json, Constraint
from mario.metadata import metadata_from_json
from mario.query_builder import ViewBasedQueryBuilder, SubsetQueryBuilder

from test.mocks import MockQueryBuilder

logger = logging.getLogger(__name__)


def test_csv_to_csv():
    dataset = dataset_from_json(os.path.join('test', 'dataset.json'))
    metadata = metadata_from_json(os.path.join('test', 'metadata.json'))
    configuration = Configuration(
        file_path=os.path.join('test', 'orders.csv')
    )
    extractor = DataExtractor(
        dataset_specification=dataset,
        metadata=metadata,
        configuration=configuration
    )
    extractor.validate_data()
    with tempfile.NamedTemporaryFile(suffix='.csv') as file:
        file.close()
        extractor.save_data_as_csv(file_path=file.name)


def test_csv_to_hyper():
    dataset = dataset_from_json(os.path.join('test', 'dataset.json'))
    metadata = metadata_from_json(os.path.join('test', 'metadata.json'))
    configuration = Configuration(
        file_path=os.path.join('test', 'orders.csv')
    )
    extractor = DataExtractor(
        dataset_specification=dataset,
        metadata=metadata,
        configuration=configuration
    )
    with tempfile.NamedTemporaryFile(suffix='.hyper') as file:
        file.close()
        extractor.save_data_as_hyper(file_path=file.name)


def test_hyper_with_nulls_to_csv():
    dataset = dataset_from_json(os.path.join('test', 'dataset.json'))
    metadata = metadata_from_json(os.path.join('test', 'metadata.json'))
    configuration = Configuration(
        file_path=os.path.join('test', 'orders_with_nulls.hyper')
    )
    extractor = DataExtractor(
        dataset_specification=dataset,
        metadata=metadata,
        configuration=configuration
    )
    with pytest.raises(ValueError):
        extractor.validate_data(allow_nulls=False)
    with tempfile.NamedTemporaryFile(suffix='.csv') as file:
        file.close()
        extractor.save_data_as_csv(file_path=file.name)


def test_hyper_without_nulls_to_csv():
    dataset = dataset_from_json(os.path.join('test', 'dataset.json'))
    metadata = metadata_from_json(os.path.join('test', 'metadata.json'))
    configuration = Configuration(
        file_path=os.path.join('test', 'orders.hyper')
    )
    extractor = DataExtractor(
        dataset_specification=dataset,
        metadata=metadata,
        configuration=configuration
    )
    extractor.validate_data(allow_nulls=False)
    with tempfile.NamedTemporaryFile(suffix='.csv') as file:
        file.close()
        extractor.save_data_as_csv(file_path=file.name)


def test_stream_sql_to_csv():
    # Skip this test if we don't have a connection string
    if not os.environ.get('CONNECTION_STRING'):
        pytest.skip("Skipping SQL test as no database configured")

    dataset = dataset_from_json(os.path.join('test', 'dataset.json'))
    metadata = metadata_from_json(os.path.join('test', 'metadata.json'))
    configuration = Configuration(
        connection_string=os.environ.get('CONNECTION_STRING'),
        schema="dev",
        view="superstore",
        query_builder=ViewBasedQueryBuilder
    )
    extractor = StreamingDataExtractor(
        dataset_specification=dataset,
        metadata=metadata,
        configuration=configuration
    )
    file = tempfile.NamedTemporaryFile(suffix='.csv')
    extractor.stream_sql_to_csv(file_path=file.name, chunk_size=1000)
    df = pd.read_csv(file)
    assert len(df) == 10194


def test_stream_sql_to_csv_with_compression():
    # Skip this test if we don't have a connection string
    if not os.environ.get('CONNECTION_STRING'):
        pytest.skip("Skipping SQL test as no database configured")

    dataset = dataset_from_json(os.path.join('test', 'dataset.json'))
    metadata = metadata_from_json(os.path.join('test', 'metadata.json'))
    configuration = Configuration(
        connection_string=os.environ.get('CONNECTION_STRING'),
        schema="dev",
        view="superstore",
        query_builder=ViewBasedQueryBuilder
    )
    extractor = StreamingDataExtractor(
        dataset_specification=dataset,
        metadata=metadata,
        configuration=configuration
    )
    file = tempfile.NamedTemporaryFile(suffix='.csv')
    gzip_path = extractor.stream_sql_to_csv(file_path=file.name, chunk_size=1000, compress_using_gzip=True)
    df = pd.read_csv(gzip_path)
    assert len(df) == 10194


def test_stream_sql_to_hyper():
    # Skip this test if we don't have a connection string
    if not os.environ.get('CONNECTION_STRING'):
        pytest.skip("Skipping SQL test as no database configured")

    dataset = dataset_from_json(os.path.join('test', 'dataset.json'))
    metadata = metadata_from_json(os.path.join('test', 'metadata.json'))
    configuration = Configuration(
        connection_string=os.environ.get('CONNECTION_STRING'),
        schema="dev",
        view="superstore",
        query_builder=ViewBasedQueryBuilder
    )
    extractor = StreamingDataExtractor(
        dataset_specification=dataset,
        metadata=metadata,
        configuration=configuration
    )
    folder = tempfile.TemporaryDirectory()
    file = os.path.join(folder.name, 'data.hyper')
    extractor.stream_sql_to_hyper(file_path=file, chunk_size=1000)
    import pantab
    from tableauhyperapi import TableName
    df = pantab.frame_from_hyper(source=file, table=TableName('Extract', 'Extract'))
    assert len(df) == 10194
    shutil.rmtree(folder.name)


def test_stream_sql_to_csv_with_validation():
    # Skip this test if we don't have a connection string
    if not os.environ.get('CONNECTION_STRING'):
        pytest.skip("Skipping SQL test as no database configured")

    dataset = dataset_from_json(os.path.join('test', 'dataset.json'))
    metadata = metadata_from_json(os.path.join('test', 'metadata.json'))
    # Restrict ship modes so that validation fails
    metadata.get_metadata('Ship Mode').set_property('domain', ["First Class", "Second Class"])
    configuration = Configuration(
        connection_string=os.environ.get('CONNECTION_STRING'),
        schema="dev",
        view="superstore",
        query_builder=ViewBasedQueryBuilder
    )
    extractor = StreamingDataExtractor(
        dataset_specification=dataset,
        metadata=metadata,
        configuration=configuration
    )
    file = tempfile.NamedTemporaryFile(suffix='.csv')
    with pytest.raises(ValueError):
        extractor.stream_sql_to_csv(file_path=file.name, validate=True, chunk_size=1000)


def test_stream_sql_to_csv_with_minimisation():
    # Skip this test if we don't have a connection string
    if not os.environ.get('CONNECTION_STRING'):
        pytest.skip("Skipping SQL test as no database configured")

    dataset = dataset_from_json(os.path.join('test', 'dataset.json'))
    metadata = metadata_from_json(os.path.join('test', 'metadata.json'))
    # Remove ship mode
    dataset.dimensions.remove('Ship Mode')
    configuration = Configuration(
        connection_string=os.environ.get('CONNECTION_STRING'),
        schema="dev",
        view="superstore",
        query_builder=ViewBasedQueryBuilder
    )
    extractor = StreamingDataExtractor(
        dataset_specification=dataset,
        metadata=metadata,
        configuration=configuration
    )
    file = tempfile.NamedTemporaryFile(suffix='.csv')
    extractor.stream_sql_to_csv(file_path=file.name, validate=False, minimise=True, chunk_size=1000)
    df = pd.read_csv(file.name)
    assert 'Ship Mode' not in df.columns


def test_column_mapping():
    # Skip this test if we don't have a connection string
    if not os.environ.get('CONNECTION_STRING'):
        pytest.skip("Skipping SQL test as no database configured")

    dataset = dataset_from_json(os.path.join('test', 'dataset.json'))
    # Rename 'region' to area
    dataset.dimensions.remove('Region')
    dataset.dimensions.append('Area')
    metadata = metadata_from_json(os.path.join('test', 'metadata.json'))
    meta = metadata.get_metadata('Region')
    meta.name = 'Area'
    meta.set_property('output_name', 'Region')
    assert metadata.get_metadata('Area') is not None
    assert metadata.get_metadata('Region') is None
    assert 'Area' in dataset.dimensions
    assert 'Region' not in dataset.dimensions

    configuration = Configuration(
        connection_string=os.environ.get('CONNECTION_STRING'),
        schema="dev",
        view="superstore",
        query_builder=ViewBasedQueryBuilder
    )
    extractor = StreamingDataExtractor(
        dataset_specification=dataset,
        metadata=metadata,
        configuration=configuration
    )
    file = tempfile.NamedTemporaryFile(suffix='.csv')
    extractor.stream_sql_to_csv(
        file_path=file.name,
        validate=True,
        chunk_size=1000
    )
    df = pd.read_csv(file.name)
    assert 'Region' in df.columns


def test_stream_to_csv_using_bcp():
    # Skip this test if we don't have a connection string
    if not os.environ.get('CONNECTION_STRING'):
        pytest.skip("Skipping SQL test as no database configured")
    if not os.environ.get('BCP'):
        pytest.skip('BCP not available')
    conn = os.environ.get('CONNECTION_STRING')

    dataset = dataset_from_json(os.path.join('test', 'dataset.json'))
    metadata = metadata_from_json(os.path.join('test', 'metadata.json'))
    configuration = Configuration(
        connection_string=conn,
        schema='dbo',
        query_builder=MockQueryBuilder
    )
    extractor = StreamingDataExtractor(
        dataset_specification=dataset,
        metadata=metadata,
        configuration=configuration
    )
    file = tempfile.NamedTemporaryFile(suffix='.csv')
    file.close()  # dumb Windows hack
    extractor.stream_sql_to_csv_using_bcp(
        table_name='v_mario_test',
        output_file_path=file.name,
        database_name=os.environ.get('DATABASE'),
        use_view=True,
        server_url=os.environ.get('SERVER'),
        delete_when_finished=True
    )


def test_hyper_totals():
    dataset = dataset_from_json(os.path.join('test', 'dataset.json'))
    metadata = metadata_from_json(os.path.join('test', 'metadata.json'))
    configuration = Configuration(
        file_path=os.path.join('test', 'orders.hyper')
    )
    extractor = DataExtractor(
        dataset_specification=dataset,
        metadata=metadata,
        configuration=configuration
    )
    extractor.validate_data()
    assert extractor.get_total() == 2326534.3543


def test_csv_totals():
    dataset = dataset_from_json(os.path.join('test', 'dataset.json'))
    metadata = metadata_from_json(os.path.join('test', 'metadata.json'))
    configuration = Configuration(
        file_path=os.path.join('test', 'orders.csv')
    )
    extractor = DataExtractor(
        dataset_specification=dataset,
        metadata=metadata,
        configuration=configuration
    )
    extractor.validate_data()
    assert extractor.get_total() == 2326534.3543
    assert extractor.get_total(measure='Sales') == 2326534.3543
    assert round(extractor.get_total(measure='Profit'), 4) == 292296.8146


def test_csv_total_profit():
    dataset = dataset_from_json(os.path.join('test', 'dataset.json'))
    dataset.measures = ['Profit']
    metadata = metadata_from_json(os.path.join('test', 'metadata.json'))
    configuration = Configuration(
        file_path=os.path.join('test', 'orders.csv')
    )
    extractor = DataExtractor(
        dataset_specification=dataset,
        metadata=metadata,
        configuration=configuration
    )
    extractor.validate_data()
    assert round(extractor.get_total(), 4) == 292296.8146


def test_csv_total_no_measures():
    dataset = dataset_from_json(os.path.join('test', 'dataset.json'))
    dataset.measures = []
    metadata = metadata_from_json(os.path.join('test', 'metadata.json'))
    configuration = Configuration(
        file_path=os.path.join('test', 'orders.csv')
    )
    extractor = DataExtractor(
        dataset_specification=dataset,
        metadata=metadata,
        configuration=configuration
    )
    extractor.validate_data()
    assert extractor.get_total() == 10194
    with pytest.raises(ValueError):
        extractor.get_total(measure='Sales')


def test_stream_sql_subset_to_csv_with_total():
    # Skip this test if we don't have a connection string
    if not os.environ.get('CONNECTION_STRING'):
        pytest.skip("Skipping SQL test as no database configured")

    dataset = dataset_from_json(os.path.join('test', 'dataset.json'))
    dataset.measures = ['Sales', 'Profit']
    dataset.dimensions = ['Product Name']
    metadata = metadata_from_json(os.path.join('test', 'metadata.json'))
    configuration = Configuration(
        connection_string=os.environ.get('CONNECTION_STRING'),
        schema="dev",
        view="superstore",
        query_builder=SubsetQueryBuilder
    )
    extractor = StreamingDataExtractor(
        dataset_specification=dataset,
        metadata=metadata,
        configuration=configuration
    )
    file = tempfile.NamedTemporaryFile(suffix='.csv')
    extractor.stream_sql_to_csv(file_path=file.name, chunk_size=1000)
    df = pd.read_csv(file)
    assert len(df) == 1849
    assert round(extractor.get_total(), 4) == 2326534.3543
    assert round(extractor.get_total(measure='Sales'), 4) == 2326534.3543
    assert round(extractor.get_total(measure='Profit'), 4) == 292296.8146


def test_stream_sql_view_to_csv_with_total():
    # Skip this test if we don't have a connection string
    if not os.environ.get('CONNECTION_STRING'):
        pytest.skip("Skipping SQL test as no database configured")

    dataset = dataset_from_json(os.path.join('test', 'dataset.json'))
    dataset.measures = ['Sales']
    dataset.dimensions = ['Product Name']
    metadata = metadata_from_json(os.path.join('test', 'metadata.json'))
    configuration = Configuration(
        connection_string=os.environ.get('CONNECTION_STRING'),
        schema="dev",
        view="superstore",
        query_builder=ViewBasedQueryBuilder
    )
    extractor = StreamingDataExtractor(
        dataset_specification=dataset,
        metadata=metadata,
        configuration=configuration
    )
    file = tempfile.NamedTemporaryFile(suffix='.csv')
    extractor.stream_sql_to_csv(file_path=file.name, chunk_size=1000)
    total = extractor.get_total()
    df = pd.read_csv(file)
    assert len(df) == 10194
    assert round(total, 2) == 2326534.35


def test_validate_data_on_streaming_extractor():
    # Skip this test if we don't have a connection string
    if not os.environ.get('CONNECTION_STRING'):
        pytest.skip("Skipping SQL test as no database configured")

    dataset = dataset_from_json(os.path.join('test', 'dataset.json'))
    metadata = metadata_from_json(os.path.join('test', 'metadata.json'))
    # Restrict ship modes so that validation fails
    metadata.get_metadata('Ship Mode').set_property('domain', ["First Class", "Second Class"])
    configuration = Configuration(
        connection_string=os.environ.get('CONNECTION_STRING'),
        schema="dev",
        view="superstore",
        query_builder=ViewBasedQueryBuilder
    )
    extractor = StreamingDataExtractor(
        dataset_specification=dataset,
        metadata=metadata,
        configuration=configuration
    )
    with pytest.raises(ValueError):
        extractor.validate_data()


def test_dataframe_extractor():
    df = pd.read_csv(os.path.join('test', 'orders.csv'))
    dataset = dataset_from_json(os.path.join('test', 'dataset.json'))
    metadata = metadata_from_json(os.path.join('test', 'metadata.json'))
    extractor = DataFrameExtractor(
        dataset_specification=dataset,
        metadata=metadata,
        dataframe=df
    )
    assert extractor.validate_data()
    assert extractor.get_total() == 2326534.3543


def test_partitioning_extractor():
    # Skip this test if we don't have a connection string
    if not os.environ.get('CONNECTION_STRING'):
        pytest.skip("Skipping SQL test as no database configured")

    dataset = dataset_from_json(os.path.join('test', 'dataset.json'))
    constraint = Constraint()
    constraint.item = 'City'
    constraint.allowed_values = ['Houston', 'Laredo', 'Springfield']
    dataset.constraints.append(constraint)
    metadata = metadata_from_json(os.path.join('test', 'metadata.json'))
    configuration = Configuration(
        connection_string=os.environ.get('CONNECTION_STRING'),
        schema="dev",
        view="superstore",
        query_builder=SubsetQueryBuilder
    )
    extractor = PartitioningExtractor(
        configuration=configuration,
        dataset_specification=dataset,
        metadata=metadata,
        partition_column='City'
    )
    extractor.__load_from_sql_using_partition__(partition_value='Houston')
    df = extractor._data
    assert len(df) == 377

    extractor._data = None
    extractor.__load_from_sql__()
    df = extractor._data
    assert df['City'].nunique() == 3


def test_partitioning_extractor_streaming():
    # Skip this test if we don't have a connection string
    if not os.environ.get('CONNECTION_STRING'):
        pytest.skip("Skipping SQL test as no database configured")

    dataset = dataset_from_json(os.path.join('test', 'dataset.json'))
    constraint = Constraint()
    constraint.item = 'City'
    constraint.allowed_values = ['Houston', 'Laredo', 'Springfield']
    dataset.constraints.append(constraint)
    metadata = metadata_from_json(os.path.join('test', 'metadata.json'))
    configuration = Configuration(
        connection_string=os.environ.get('CONNECTION_STRING'),
        schema="dev",
        view="superstore",
        query_builder=SubsetQueryBuilder
    )
    extractor = PartitioningExtractor(
        configuration=configuration,
        dataset_specification=dataset,
        metadata=metadata,
        partition_column='City'
    )

    path = os.path.join('output', 'test_partitioning_extractor_streaming')
    file = os.path.join(path, 'test.hyper')
    csv_file = os.path.join(path, 'test.csv')

    os.makedirs(path, exist_ok=True)

    extractor.stream_sql_to_hyper(
        file_path=file,
    )
    extractor.stream_sql_to_csv(
        file_path=csv_file,
        compress_using_gzip=True
    )


def test_hyper_total_measure():
    dataset = dataset_from_json(os.path.join('test', 'dataset.json'))
    dataset.measures = []
    metadata = metadata_from_json(os.path.join('test', 'metadata.json'))
    configuration = Configuration(
        file_path=os.path.join('test', 'orders.hyper')
    )
    extractor = HyperFile(
        dataset_specification=dataset,
        metadata=metadata,
        configuration=configuration
    )
    assert extractor.get_total() == 10194
    assert extractor.get_total(measure='Sales') == 2326534.354299952




import datetime
import shutil

import pytest

from mario.dataset_splitter import DatasetSplitter
import os
import csv


def count_rows_in_csv(file_path):
    with open(file_path, 'r') as file:
        reader = csv.reader(file)
        row_count = sum(1 for row in reader) - 1  # Subtract 1 to exclude the header row
    return row_count


def test_dataset_splitter_output_dir_has_subdirectories():
    output_dir = os.path.join('output', 'test_dataset_splitter_output_dir_has_subdirectories')
    os.makedirs(os.path.join(output_dir, 'fruit', 'banana'), exist_ok=True)
    with pytest.raises(ValueError):
        DatasetSplitter(
            field='Region',
            source_path=os.path.join('test', 'test_split_source'),
            output_path=output_dir
        )
    shutil.rmtree(output_dir)


def test_dataset_splitter_output_dir_has_deep_subdirectories():
    output_dir = os.path.join('output', 'test_dataset_splitter_output_dir_has_deep_subdirectories')
    os.makedirs(os.path.join(output_dir, 'fruit', 'banana', 'yellow'), exist_ok=True)
    with pytest.raises(ValueError):
        DatasetSplitter(
            field='Region',
            source_path=os.path.join('test', 'test_split_source'),
            output_path=output_dir
        )
    shutil.rmtree(output_dir)


def test_dataset_splitter_source_dir_does_not_exist():
    output_dir = os.path.join('output', 'test_dataset_splitter_source_dir_does_not_exist')
    os.makedirs(output_dir, exist_ok=True)
    with pytest.raises(FileNotFoundError):
        DatasetSplitter(
            field='Region',
            source_path=os.path.join('test', 'no_such_folder'),
            output_path=output_dir
        )
    shutil.rmtree(output_dir)


def test_dataset_splitter_source_dir_has_subdirectories():
    source_dir = os.path.join('output', 'test_dataset_splitter_source_dir_has_subdirectories_source')
    os.makedirs(os.path.join(source_dir, 'For_Supply'), exist_ok=True)
    output_dir = os.path.join('output', 'test_dataset_splitter_source_dir_has_subdirectories_output')
    os.makedirs(output_dir, exist_ok=True)
    with pytest.raises(ValueError):
        DatasetSplitter(
            field='Region',
            source_path=source_dir,
            output_path=output_dir
        )
    shutil.rmtree(source_dir)
    shutil.rmtree(output_dir)


def test_dataset_splitter_source_and_output_are_the_same():
    source_dir = os.path.join('output', 'test_dataset_splitter_source_and_output_are_the_same')
    output_dir = os.path.join('output', 'test_dataset_splitter_source_and_output_are_the_same')
    os.makedirs(output_dir, exist_ok=True)
    with pytest.raises(ValueError):
        DatasetSplitter(
            field='Region',
            source_path=source_dir,
            output_path=output_dir
        )

    shutil.rmtree(output_dir)


def test_dataset_splitter_no_files_to_split():
    source_dir = os.path.join('output', 'test_dataset_splitter_no_files_to_split_source')
    output_dir = os.path.join('output', 'test_dataset_splitter_no_files_to_split_output')
    os.makedirs(output_dir, exist_ok=True)
    os.makedirs(source_dir, exist_ok=True)
    shutil.copyfile(os.path.join('test', 'README.md'), os.path.join(output_dir, 'README.md'))
    with pytest.raises(ValueError) as e:
        data_splitter = DatasetSplitter(
            field='Region',
            source_path=source_dir,
            output_path=output_dir
        )
        data_splitter.split_files()
    shutil.rmtree(source_dir)
    shutil.rmtree(output_dir)


def test_dataset_splitter():
    # Get the total rows from the test CSV
    total_rows_original = count_rows_in_csv(os.path.join('test', 'test_split_source', 'orders.csv'))

    dataset_splitter = DatasetSplitter(
        field='Region',
        source_path=os.path.join('test', 'test_split_source'),
        output_path=os.path.join('output', 'test_dataset_splitter')
    )
    dataset_splitter.split_files()
    dataset_splitter.copy_other_files()
    assert 'Central' in os.listdir(os.path.join('output', 'test_dataset_splitter'))
    assert 'orders.csv' in os.listdir(os.path.join('output', 'test_dataset_splitter', 'Central'))
    assert 'orders.xlsx' in os.listdir(os.path.join('output', 'test_dataset_splitter', 'Central'))
    assert 'README.txt' in os.listdir(os.path.join('output', 'test_dataset_splitter', 'Central'))

    assert len(os.listdir(os.path.join('output', 'test_dataset_splitter'))) == 4

    total_rows = 0
    for region in ['Central', 'East', 'West', 'South']:
        total_rows += count_rows_in_csv(os.path.join('output', 'test_dataset_splitter', region, 'orders.csv'))
    assert total_rows_original == total_rows


def test_dataset_splitter_pivot():

    dataset_splitter = DatasetSplitter(
        field='Category',
        source_path=os.path.join('test', 'test_split_pivot'),
        output_path=os.path.join('output', 'test_dataset_splitter_pivot')
    )
    dataset_splitter.split_files()
    dataset_splitter.copy_other_files()
    assert 'Furniture' in os.listdir(os.path.join('output', 'test_dataset_splitter_pivot'))
    assert len(os.listdir(os.path.join('output', 'test_dataset_splitter_pivot'))) == 3


def test_dataset_splitter_info():

    dataset_splitter = DatasetSplitter(
        field='Category',
        source_path=os.path.join('test', 'test_split_with_info'),
        output_path=os.path.join('output', 'test_dataset_splitter_info')
    )
    dataset_splitter.split_files()
    dataset_splitter.copy_other_files()
    assert 'Furniture' in os.listdir(os.path.join('output', 'test_dataset_splitter_info'))
    assert len(os.listdir(os.path.join('output', 'test_dataset_splitter_info'))) == 3
    assert 'info.xlsx' in os.listdir(os.path.join('output', 'test_dataset_splitter_info', 'Furniture'))


def test_dataset_splitter_gz():

    dataset_splitter = DatasetSplitter(
        field='Category',
        source_path=os.path.join('test', 'test_split_gz'),
        output_path=os.path.join('output', 'test_dataset_splitter_gz')
    )
    dataset_splitter.split_files()
    dataset_splitter.copy_other_files()
    assert 'Furniture' in os.listdir(os.path.join('output', 'test_dataset_splitter_gz'))
    assert len(os.listdir(os.path.join('output', 'test_dataset_splitter_gz'))) == 3
    assert 'orders.csv' in os.listdir(os.path.join('output', 'test_dataset_splitter_info', 'Furniture'))


def test_dataset_splitter_parallel():
    """
    This mimics a workflow that takes the elements
    of the splitter and runs them in separate processes.
    :return: None
    """

    # Get the total rows from the test CSV
    total_rows_original = count_rows_in_csv(os.path.join('test', 'test_split_source', 'orders.csv'))

    dataset_splitter = DatasetSplitter(
        field='Region',
        source_path=os.path.join('test', 'test_split_source'),
        output_path=os.path.join('output', 'test_dataset_splitter_parallel')
    )
    files_to_split, files_to_copy = dataset_splitter.get_files_to_split_or_copy()
    print("Files to split", files_to_split)
    print("Files to copy", files_to_copy)
    for file in files_to_split:
        if file.endswith('.xlsx'):
            print("Splitting Excel file", file)
            for value in dataset_splitter.get_excel_values(file):
                print("Splitting Excel file by value", value)
                dataset_splitter.split_excel_by_value(file, value)
        else:
            print("Splitting CSV file", file)
            dataset_splitter.split_file(file)
    for file in files_to_copy:
        dataset_splitter.copy_other_file(file)
    print("Split files", os.listdir(os.path.join('output', 'test_dataset_splitter_parallel', 'Central')))
    assert 'Central' in os.listdir(os.path.join('output', 'test_dataset_splitter_parallel'))
    assert 'orders.csv' in os.listdir(os.path.join('output', 'test_dataset_splitter_parallel', 'Central'))
    assert 'orders.xlsx' in os.listdir(os.path.join('output', 'test_dataset_splitter_parallel', 'Central'))
    assert 'README.txt' in os.listdir(os.path.join('output', 'test_dataset_splitter_parallel', 'Central'))

    assert len(os.listdir(os.path.join('output', 'test_dataset_splitter_parallel'))) == 4

    total_rows = 0
    for region in ['Central', 'East', 'West', 'South']:
        total_rows += count_rows_in_csv(os.path.join('output', 'test_dataset_splitter_parallel', region, 'orders.csv'))
    assert total_rows_original == total_rows


def test_dataset_splitter_builder():
    print('start', datetime.datetime.now().isoformat())
    dataset_splitter = DatasetSplitter(
        field='Ship Mode',
        source_path=os.path.join('test', 'test_dataset_splitter_builder'),
        output_path=os.path.join('output', 'test_dataset_splitter_builder'),
        use_excel_builder=True,
        template='excel_template.xlsx'
    )
    dataset_splitter.split_files()
    dataset_splitter.copy_other_files()
    assert 'First Class' in os.listdir(os.path.join('output', 'test_dataset_splitter_builder'))
    assert len(os.listdir(os.path.join('output', 'test_dataset_splitter_builder'))) == 4
    assert 'orders.xlsx' in os.listdir(os.path.join('output', 'test_dataset_splitter_builder', 'First Class'))
    print('end', datetime.datetime.now().isoformat())



"""
A utility for partitioning dataset files by a particular field
"""
import os
import shutil
import csv
import logging

logger = logging.getLogger(__name__)


class DatasetSplitter:

    def __init__(self, field: str, source_path: str, output_path: str, use_excel_builder=False, template=None):
        self.field = field
        self.source_path = source_path
        self.output_path = output_path
        self.other_files = []
        self.use_excel_builder = use_excel_builder
        self.template = template

        if self.use_excel_builder is True and (template is None or not os.path.exists(template)):
            raise ValueError("You must specify a valid template file to use the Excel Builder")

        # If the source path doesn't exist, raise an error
        if not os.path.exists(self.source_path):
            raise FileNotFoundError("Source folder doesn't exist.")

        # If the source path and output path are the same, raise an error
        if self.source_path == self.output_path:
            raise ValueError("Source and output folders can't be the same.")

        # If the output path contains subdirectories, raise an error
        if os.path.exists(self.output_path):
            if any(os.path.relpath(root, self.output_path).count(os.sep) > 0 for root, dirs, _ in os.walk(self.output_path)):
                raise ValueError("Output folder contains subdirectories so is probably not something you "
                                 "want to delete and overwrite.")

        # If the source path contains subdirectories, raise an error
        if next(os.walk(self.source_path))[1]:
            raise ValueError("Source folder contains subdirectories so is probably not something you "
                             "want to split as the subdirectories aren't going to be processed.")

        # Delete the output path if it exists
        if os.path.exists(self.output_path):
            logger.info(f"Removing output path at {self.output_path}")
            shutil.rmtree(self.output_path)

        os.makedirs(self.output_path)

    def get_output_path(self, file, field_value):
        return os.path.join(self.output_path, field_value, file)

    def get_files_to_split_or_copy(self):
        # Iterate over each file in the source folder
        files_to_split = []
        files_to_copy = []
        for file in os.listdir(self.source_path):
            if os.path.isfile(os.path.join(self.source_path, file)):
                if file.endswith('.xlsx'):
                    try:
                        self.get_excel_values(file)
                        files_to_split.append(file)
                        logger.debug(f"Excel file {file} is to be split")
                    except Exception:
                        logger.debug(f"Excel file {file} has no values for field")
                        files_to_copy.append(file)
                elif file.endswith('.csv'):
                    files_to_split.append(file)
                elif file.endswith('.csv.gz'):
                    files_to_split.append(file)
                else:
                    files_to_copy.append(file)
        if len(files_to_split) == 0:
            raise ValueError("No files were split")
        return files_to_split, files_to_copy

    def split_file(self, file):
        if file.endswith('.xlsx'):
            if self.use_excel_builder:
                self.split_excel_using_builder(file)
            else:
                self.split_excel(file)
        elif file.endswith('.csv'):
            self.split_csv(file)
        elif file.endswith('.csv.gz'):
            self.split_gzipped_csv(file)

    def split_files(self):
        files_to_split, files_to_copy = self.get_files_to_split_or_copy()
        for file in files_to_split:
            self.split_file(file)

    def copy_other_file(self, file):
        logger.info(f"Copying file {file} to output directory")
        for output_directory in os.listdir(self.output_path):
            if os.path.isdir(os.path.join(self.output_path, output_directory)):
                shutil.copyfile(
                    os.path.join(self.source_path, file),
                    os.path.join(self.output_path, output_directory, file))

    def copy_other_files(self) -> None:
        """
        Iterates over the source folder, and copies each non-data file into
        each 'split' output directory present
        :return: None
        """
        files_to_split, files_to_copy = self.get_files_to_split_or_copy()
        for file in files_to_copy:
            self.copy_other_file(file)

    def process_batch(self, batch, column_name, file_handles, file_name):
        for row in batch:
            value = row[column_name]
            if value not in file_handles:
                os.makedirs(os.path.join(self.output_path, value), exist_ok=True)
                file_path = self.get_output_path(field_value=value, file=file_name)
                file_handles[value] = open(file_path, 'w', newline='')
                writer = csv.DictWriter(file_handles[value], fieldnames=row.keys())
                writer.writeheader()
            writer = csv.DictWriter(file_handles[value], fieldnames=row.keys())
            writer.writerow(row)

    def split_gzipped_csv(self, file_name: str, batch_size=10000):
        import gzip

        logger.info(f"Splitting gzipped CSV file {file_name}")

        # Dict for holding open file handles
        file_handles = {}

        # Path to the CSV to split
        file_path = os.path.join(self.source_path, file_name)

        output_file_name = file_name.rstrip('.gz')

        with gzip.open(file_path, 'rt', newline='') as infile:
            reader = csv.DictReader(infile)

            # Process the input file in batches
            batch = []
            for row in reader:
                batch.append(row)
                if len(batch) >= batch_size:
                    self.process_batch(batch, self.field, file_handles, output_file_name)
                    batch = []

            # Process any remaining rows in the last batch
            if batch:
                self.process_batch(batch, self.field, file_handles, output_file_name)

        # Close all open file handles
        for handle in file_handles.values():
            handle.close()

    def split_csv(self, file_name: str, batch_size=10000, compression=None):

        logger.info(f"Splitting CSV file {file_name}")

        # Dict for holding open file handles
        file_handles = {}

        # Path to the CSV to split
        file_path = os.path.join(self.source_path, file_name)

        # Open the input CSV file for reading
        with open(file_path, 'r') as infile:
            reader = csv.DictReader(infile)

            # Process the input file in batches
            batch = []
            for row in reader:
                batch.append(row)
                if len(batch) >= batch_size:
                    self.process_batch(batch, self.field, file_handles, file_name)
                    batch = []

            # Process any remaining rows in the last batch
            if batch:
                self.process_batch(batch, self.field, file_handles, file_name)

        # Close all open file handles
        for handle in file_handles.values():
            handle.close()

    def split_excel_using_builder(self, file_name: str) -> None:
        """
        Splits an Excel file by using the ExcelBuilder to create
        a new workbook containing both a Data sheet and a Pivot sheet
        :param file_name: the name of the Excel file to split
        :return: None
        """
        from mario.excel_builder import ExcelBuilder
        from mario.data_extractor import DataFrameExtractor
        from mario.dataset_specification import DatasetSpecification
        from mario.metadata import Metadata, Item
        from mario.excel_pivot_utils import prepend_sheet_to_workbook, get_worksheet_containing_field, get_info_sheets
        import pandas as pd
        from openpyxl import load_workbook

        file_path = os.path.join(self.source_path, file_name)

        # Create a dataset based on the contents of the 'Data' sheet in the Excel workbook. We find this
        # by locating a 'data sheet' that has the field name in the header. At the same time we identify
        # any 'info' sheets that don't have any data as we want to copy those as-is
        workbook = load_workbook(file_path)
        data_sheet = get_worksheet_containing_field(workbook, self.field)
        info_sheets = get_info_sheets(workbook, self.field)

        if data_sheet is None:
            logger.warning("Encountered an Excel file with no data sheet; treating as an 'other' file")
            self.other_files.append(file_name)
            return

        # Read in the data and get unique values for the field we want to split by
        df = pd.read_excel(file_path, sheet_name=data_sheet,  dtype={self.field: object}, header=0)
        values = df[self.field].unique()

        # Generate some basic metadata
        ds = DatasetSpecification()
        ds.dimensions = list(df.columns)
        ds.measures = []
        md = Metadata()
        for field in df.columns:
            item = Item()
            item.name = field
            md.add_item(item)

        # Create a data extractor instance for the DF
        dx = DataFrameExtractor(
            dataset_specification=ds,
            metadata=md,
            dataframe=df
        )

        # Create an ExcelBuilder instance for creating excel files
        builder = ExcelBuilder(
            output_file_path='',
            template_path=self.template,
            dataset_specification=ds,
            metadata=md,
            data_extractor=dx
        )

        # For each unique value, subset the data and create the output
        for value in values:
            logger.info(f"Splitting {file_name} for value {value}")
            output_folder = os.path.join(self.output_path, value)
            os.makedirs(output_folder, exist_ok=True)
            output_path = os.path.join(output_folder, file_name)
            builder.filepath = output_path
            subset = df[df[self.field] == value]
            dx._data = subset
            builder.create_workbook(create_notes_page=False)

            # prepend any info sheets, if present
            for info_sheet in info_sheets:
                prepend_sheet_to_workbook(
                    source_workbook_file=file_path,
                    target_workbook_file=output_path,
                    sheet_name=info_sheet
                )

    def get_excel_values(self, file_name: str):
        from mario.excel_pivot_utils import get_unique_values_for_workbook
        file_path = os.path.join(self.source_path, file_name)
        return get_unique_values_for_workbook(file_path=file_path, field=self.field)

    def split_excel_by_value(self, file_name, value: str):
        """
        Splits an Excel file by a single value.
        :param file_name: the name of the Excel file (i.e. relative to the source path)
        :param value: the value to split by
        :return: the path to the split file for the value given e.g. output/Central/filename.xlsx
        """
        import pandas as pd
        from openpyxl import load_workbook
        from mario.excel_pivot_utils import set_excel_active_sheet

        file_path = os.path.join(self.source_path, file_name)

        logger.info(f"Splitting for value {value}")
        sheet_names = pd.ExcelFile(file_path).sheet_names
        split_output_path = os.path.join(self.output_path, str(value))
        split_workbook_path = os.path.join(split_output_path, file_name)
        os.makedirs(split_output_path, exist_ok=True)
        shutil.copyfile(src=file_path, dst=split_workbook_path)

        # Load the workbook once per 'run'
        workbook = load_workbook(split_workbook_path)
        for sheet_name in sheet_names:
            if len(workbook.get_sheet_by_name(sheet_name)._pivots) > 0:
                logger.info(f"Splitting excel pivot")
                self.split_excel_pivot(
                    workbook=workbook,
                    file_path=split_workbook_path,
                    sheet_name=sheet_name,
                    value=value
                )
            else:
                logger.info(f"Splitting excel sheet")
                self.split_excel_table(
                    workbook=workbook,
                    file_path=split_workbook_path,
                    sheet_name=sheet_name,
                    value=value
                )
        # Close the workbook explicitly once we're finished with it

        workbook.close()
        # Clears the active sheet so we don't have any grouped tabs. This
        # also seems to have a side effect of refreshing the pivot cache
        set_excel_active_sheet(split_workbook_path)

        return split_workbook_path

    def split_excel(self, file_name: str):
        """
        Convenience method for splitting an Excel file. Call the split methods
        directly to process in parallel.
        :param file_name:
        :return:
        """
        from mario.excel_pivot_utils import set_excel_active_sheet
        # Get the values
        try:
            values = self.get_excel_values(file_name=file_name)
        except ValueError:
            logger.warning("Encountered an Excel file with no data; treating as an 'other' file")
            self.other_files.append(file_name)
            return False

        # Create split workbooks
        for value in values:
            split_workbook_path = self.split_excel_by_value(file_name=file_name, value=value)
            set_excel_active_sheet(file_path=split_workbook_path)

    def split_excel_pivot(self, workbook, file_path: str, sheet_name: str, value: str):
        """
        Splits an Excel pivot sheet
        :param workbook: the workbook to split
        :param file_path: the path to write the split file to
        :param sheet_name: the name of the sheet
        :param value: the value to split by
        :return: None
        """
        from mario.excel_pivot_utils import replace_pivot_cache_with_subset
        ws = workbook[sheet_name]
        replace_pivot_cache_with_subset(ws, self.field, value)
        workbook.save(file_path)

    def split_excel_table(self, workbook, file_path: str, sheet_name: str, value):
        """
        Splits a standard worksheet (i.e. not a pivot) for the specified value
        :param workbook: the workbook to split
        :param file_path: the file path to write the split file to
        :param sheet_name: the name of the sheet
        :param value: the value to split by
        :return: None
        """

        ws = workbook[sheet_name]
        header = [cell.value for cell in ws[1]]  # Read header row

        if self.field not in header:
            return None

        # Create a new sheet for filtered data
        workbook.create_sheet(title=f"{sheet_name}_filtered")
        new_ws = workbook[f"{sheet_name}_filtered"]
        new_ws.append(header)

        # Read rows efficiently. Note that using comprehension first rather than writing direct from iter_rows
        # speeds things up by about 50% at a cost of higher memory use.
        field_index = header.index(self.field)
        filtered_rows = [row for row in ws.iter_rows(min_row=2, values_only=True) if row[field_index] == value]
        for row in filtered_rows:
            new_ws.append(row)

        # Save and remove original sheet
        workbook.remove(ws)
        new_ws.title = sheet_name
        workbook.save(file_path)



import logging
import os.path
import shutil
import tempfile
from enum import Enum


from mario.data_extractor import DataExtractor
from mario.dataset_specification import DatasetSpecification
from mario.metadata import Metadata

logger = logging.getLogger(__name__)


class Format(Enum):
    TABLEAU_PACKAGED_DATASOURCE = 'tdsx'
    POWER_BI_PACKAGE = 'pbix'
    EXCEL_PIVOT = 'xlsx'
    CSV = 'csv'
    EXCEL_INFO_SHEET = 'info'


class DatasetBuilder:

    def __init__(self,
                 dataset_specification: DatasetSpecification = None,
                 metadata: Metadata = None,
                 data: DataExtractor = None
                 ):
        self.dataset_specification = dataset_specification
        self.metadata = metadata
        self.data = data

    def validate_metadata(self):
        """ Checks whether the metadata matches the specification"""
        not_included = []
        fields = self.dataset_specification.items

        for field in fields:
            try:
                self.metadata.get_metadata(field)
            except ValueError:
                not_included.append(field)

        if len(not_included) > 0:
            logger.error(self.dataset_specification.name + ": Some metadata is missing")
            logger.error(not_included)
            return False
        else:
            logger.info(self.dataset_specification.name + ": All metadata present")
            return True

    def remove_redundant_hierarchies(self):
        """
        Modifies the metadata to remove any hierarchies that contain
        only one item.
        :return: None
        """
        for hierarchy in self.metadata.get_hierarchies():
            items = self.metadata.get_hierarchy(hierarchy)
            items = [item for item in items if item in self.dataset_specification.items]
            if len(items) == 1:
                for item in self.metadata.items:
                    if 'hierarchies' in item.properties:
                        item.set_property('hierarchies',  [h for h in item.get_property('hierarchies') if h['hierarchy'] != hierarchy])

    def build(self, output_format: Format, file_path: str, template_path: str = None):
        if output_format == Format.TABLEAU_PACKAGED_DATASOURCE:
            self.__build_tdsx__(file_path)
        elif output_format == Format.CSV:
            self.__build_csv__(file_path)
        elif output_format == Format.EXCEL_PIVOT:
            self.__build_excel_pivot__(file_path, template_path)
        elif output_format == Format.EXCEL_INFO_SHEET:
            self.__build_excel_info_sheet(file_path, template_path)
        else:
            raise NotImplementedError

    def __build_excel_info_sheet(self, file_path: str, template_path: str):
        from mario.excel_builder import ExcelBuilder
        excel_builder = ExcelBuilder(
            output_file_path=file_path,
            dataset_specification=self.dataset_specification,
            metadata=self.metadata,
            data_extractor=self.data,
            template_path=template_path
        )
        excel_builder.create_notes_only()

    def __build_excel_pivot__(self, file_path: str, template_path: str):
        from mario.excel_builder import ExcelBuilder
        if self.data.get_total() > 1000000:
            logger.warning("The dataset is larger than 1m rows; this isn't supported in Excel format")
        excel_builder = ExcelBuilder(
            output_file_path=file_path,
            dataset_specification=self.dataset_specification,
            metadata=self.metadata,
            data_extractor=self.data,
            template_path=template_path
        )
        excel_builder.create_workbook()

    def __build_csv__(self, file_path: str, compress_using_gzip=False):
        # TODO export Info sheet as well - see code in Automation2.0 and TDSA.
        self.data.save_data_as_csv(file_path=file_path, compress_using_gzip=compress_using_gzip)

    def __build_tdsx__(self, file_path: str):
        from mario.hyper_utils import get_default_table_and_schema
        from tableau_builder.json_metadata import JsonRepository
        with tempfile.TemporaryDirectory() as temp_folder:

            # Metadata - convert to internal format used for Tableau
            metadata_file_path = os.path.join(temp_folder, 'metadata.json')
            self.metadata.save(metadata_file_path)
            repository = JsonRepository(metadata_file_path)

            # Move the hyper extract
            data_file_path = os.path.join(temp_folder, 'data.hyper')
            self.data.save_data_as_hyper(file_path=data_file_path)

            # Get the table and schema name from the extract
            schema, table = get_default_table_and_schema(data_file_path)

            # Save the spec
            dataset_file_path = os.path.join(temp_folder, 'dataset.json')
            self.dataset_specification.save(dataset_file_path)

            # Build a tdsx in a temp file
            output_path = os.path.join(temp_folder, 'build')
            from tableau_builder.dataset import create_tdsx, HYPER
            create_tdsx(
                dataset_file=dataset_file_path,
                metadata_repository=repository,
                data_file=data_file_path,
                output_file=output_path,
                data_source_type=HYPER,
                table_name=table,
                schema_name=schema,
                use_metadata_groups=True
            )
            shutil.copyfile(src=output_path + '.tdsx', dst=file_path)



import logging
import shutil
import tempfile

import pandas as pd
from pandas import DataFrame

from mario.dataset_specification import DatasetSpecification
from mario.metadata import Metadata

logger = logging.getLogger(__name__)


class Configuration:
    """
    Configuration for a data extractor. This can include:
    * A connection string: an ODBC connection string for SQL extraction
    * A view name: A view as the base for the SQL code (H+, TDSA only)
    * A hook: An object or function used to access data, e.g. S3
    * A file path: A data file to load as the source of data, e.g. data.csv
    * A QueryBuilder class: instantiate and use to build a SQL query
    """

    def __init__(self,
                 connection_string: str = None,
                 hook=None,
                 view: str = None,
                 schema: str = None,
                 file_path: str = None,
                 query_builder=None,
                 user: str = None
                 ):
        self.connection_string = connection_string
        self.hook = hook
        self.view = view
        self.schema = schema
        self.file_path = file_path
        self.query_builder = query_builder
        self.user = user


class DataExtractor:

    def __init__(self,
                 configuration: Configuration,
                 dataset_specification: DatasetSpecification,
                 metadata: Metadata):
        self.configuration = configuration
        self.dataset_specification = dataset_specification
        self.metadata = metadata
        self._data = None
        self._query = None
        self._total = 0

    def __load__(self):
        if self.configuration is not None:
            if self.configuration.file_path is not None:
                if self.configuration.file_path.endswith('.csv'):
                    self.__load_from_csv__()
                elif self.configuration.file_path.endswith('.hyper'):
                    self.__load_from_hyper__()
                else:
                    raise ValueError("Unsupported file type")
            elif self.configuration.hook is not None:
                self.__load_from_hook__()
            else:
                self.__load_from_sql__()

    def __load_from_hook__(self):
        self.__build_query__()
        logger.info("Executing query using hook")
        self._data = self.configuration.hook.get_pandas_df(sql=self._query[0], parameters=self._query[1])

    def __load_from_sql__(self):
        self.__build_query__()
        logger.info("Executing query")
        from sqlalchemy import create_engine
        engine = create_engine(self.configuration.connection_string)
        self._data = pd.read_sql(sql=self._query[0], con=engine.connect(), params=self._query[1])

    def __build_query__(self):
        logger.info("Building query")
        self._query = ''
        if self.configuration.query_builder is not None:
            from mario.query_builder import QueryBuilder
            query_builder: QueryBuilder = self.configuration.query_builder(
                configuration=self.configuration,
                metadata=self.metadata,
                dataset_specification=self.dataset_specification)
            self._query = query_builder.create_query()
        else:
            raise NotImplementedError

    def __load_from_csv__(self):
        self._data = pd.read_csv(self.configuration.file_path)

    def __load_from_hyper__(self):
        import pantab
        from tableauhyperapi import TableName
        from mario.hyper_utils import get_default_table_and_schema
        schema, table = get_default_table_and_schema(hyper_path=self.configuration.file_path)
        table = TableName(schema, table)
        self._data = pantab.frame_from_hyper(
            source=self.configuration.file_path,
            table=table
        )

    def __get_column_name__(self, item: str):
        """ Returns the column name for a metadata item"""
        meta = self.metadata.get_metadata(item)
        if meta.get_property('output_name') is not None:
            return meta.get_property('output_name')
        elif meta.get_property('physical_column_name') is not None:
            return meta.get_property('physical_column_name')
        else:
            return meta.name

    def __minimise_data__(self):
        """ Minimise data so we only keep the columns in the spec """
        columns_to_keep = []
        for item in self.dataset_specification.items:
            if self.metadata.get_metadata(item) and not self.metadata.get_metadata(item).get_property('formula'):
                columns_to_keep.append(self.__get_column_name__(item))
        self._data = self._data[columns_to_keep]

    def __get_measure__(self, measure=None):
        if measure is None:
            if len(self.dataset_specification.measures) > 0:
                measure = self.dataset_specification.measures[0]
        else:
            if measure not in self.dataset_specification.measures:
                raise ValueError(f"Measure {measure} does not exist in dataset specification")
        return measure

    def get_total(self, measure=None):
        df = self.get_data_frame()
        measure = self.__get_measure__(measure)
        if measure is None:
            return len(df)
        self._total = df[measure].sum()
        return self._total

    def validate_data(self, allow_nulls=True):
        from mario.validation import DataFrameValidator
        validator = DataFrameValidator(
            self.dataset_specification,
            self.metadata,
            data=self.get_data_frame()
        )
        return validator.validate_data(allow_nulls)

    def get_data_frame(self, minimise=True) -> DataFrame:
        if self._data is None:
            self.__load__()
        if minimise:
            self.__minimise_data__()
        return self._data

    def save_query(self, file_path: str, formatted: bool = False):
        """
        Output the query used
        :param file_path:
        :param formatted: whether to format parameters inline; defaults to False
        :return:
        """
        from mario.query_builder import get_formatted_query
        if self._query is None:
            self.__build_query__()
        sql = self._query[0]
        if formatted:
            sql = get_formatted_query(self._query[0], self._query[1])
        with open(file_path, mode='w') as file:
            file.write(sql)

    def save_data_as_csv(self, file_path: str, minimise=True, compress_using_gzip=False):
        if self._data is None:
            self.__load__()
        if minimise:
            self.__minimise_data__()
        self._data.to_csv(file_path)

    def save_data_as_hyper(self, file_path: str, table: str = 'Extract', schema: str = 'Extract', minimise=True):
        import pantab
        from tableauhyperapi import TableName
        if self._data is None:
            self.__load__()
        table_name = TableName(schema, table)
        if minimise:
            self.__minimise_data__()
        pantab.frame_to_hyper(df=self._data, database=file_path, table=table_name)


class HyperFile(DataExtractor):
    """
    Wrapper for a HyperFile as a data extractor - use when no data needs to be extracted,
    and we just want to treat a hyper as a hyper with no conversion to/from dataframe
    """
    def __init__(self,
                 configuration: Configuration,
                 dataset_specification: DatasetSpecification,
                 metadata: Metadata):
        super().__init__(Configuration(), dataset_specification, metadata)
        self.configuration = configuration

    def validate_data(self, allow_nulls=True):
        from mario.validation import HyperValidator
        validator = HyperValidator(
            dataset_specification=self.dataset_specification,
            metadata=self.metadata,
            hyper_file_path=self.configuration.file_path
        )
        return validator.validate_data(allow_nulls)

    def __minimise_data__(self):
        from mario.hyper_utils import get_default_table_and_schema, drop_columns_from_hyper
        columns_to_keep = self.dataset_specification.items
        schema, table = get_default_table_and_schema(self.configuration.file_path)
        drop_columns_from_hyper(
            hyper_file_path=self.configuration.file_path,
            columns_to_keep=columns_to_keep,
            schema=schema,
            table=table
        )

    def get_total(self, measure=None):
        from mario.hyper_utils import get_row_count, get_default_table_and_schema, get_total
        schema, table = get_default_table_and_schema(self.configuration.file_path)
        if measure is None:
            return get_row_count(
                hyper_file_path=self.configuration.file_path,
                schema=schema,
                table=table
            )
        else:
            return get_total(
                hyper_file_path=self.configuration.file_path,
                schema=schema,
                table=table,
                measure=measure
            )

    def save_data_as_hyper(self, file_path: str, table: str = 'Extract', schema: str = 'Extract', minimise=False):
        if minimise:
            self.__minimise_data__()
        shutil.copyfile(self.configuration.file_path, file_path)

    def save_data_as_csv(self,
                         file_path: str,
                         minimise=True,
                         compress_using_gzip=False,
                         chunk_size=100000
                         ):
        from mario.hyper_utils import get_default_table_and_schema
        import pantab

        if compress_using_gzip:
            compression_options = dict(method='gzip')
            file_path = file_path + '.gz'
        elif file_path.endswith('.gz'):
            compression_options = dict(method='gzip')
        else:
            compression_options = None

        mode = 'w'
        header = True
        offset = 0
        schema, table = get_default_table_and_schema(self.configuration.file_path)
        sql = f"SELECT * FROM \"{schema}\".\"{table}\""

        while True:
            query = f"{sql} LIMIT {chunk_size} OFFSET {offset}"
            df_chunk = pantab.frame_from_hyper_query(self.configuration.file_path, query)
            if df_chunk.empty:
                break
            df_chunk.to_csv(file_path, index=False, mode=mode, header=header,
                            compression=compression_options)
            offset += chunk_size
            if header:
                header = False
                mode = "a"


class StreamingDataExtractor(DataExtractor):
    """
    Extension to basic data extractor with additional functions
    supporting streaming data from SQL to output formats without
    holding any data in memory using a data frame
    """
    def __init__(self,
                 configuration: Configuration,
                 dataset_specification: DatasetSpecification,
                 metadata: Metadata
                 ):
        super().__init__(configuration, dataset_specification, metadata)
        self._data = None

    def get_data_frame(self, minimise=True) -> DataFrame:
        if self._data is None:
            raise NotImplementedError("Dataframe is not available when using a streaming extractor")
        else:
            return super().get_data_frame(minimise=minimise)

    def validate_data(self, allow_nulls=True):
        if self._data is None:
            logger.warning("Calling validate_data() on a streaming extractor causes the data"
                           "to be re-streamed from source into a temporary file. If this is "
                           "not what you had in mind, first stream the data to file, and "
                           "then load the file in a DataExtractor")
            with tempfile.TemporaryFile() as file:
                self.stream_sql_to_hyper(file_path=file.name, validate=True, allow_nulls=allow_nulls)
        else:
            super().validate_data(allow_nulls=allow_nulls)

    def get_connection(self):
        from sqlalchemy import create_engine

        if self.configuration.hook is not None:
            return self.configuration.hook.get_conn()

        engine = create_engine(self.configuration.connection_string)
        connection = engine.connect().execution_options(stream_results=True)
        return connection

    def save_data_as_csv(self, file_path: str, minimise=False, compress_using_gzip=False):
        if minimise:
            raise NotImplementedError('Cannot minimise data when using streaming')
        self.stream_sql_to_csv(file_path=file_path, compress_using_gzip=compress_using_gzip)

    def save_data_as_hyper(self, file_path: str, table: str = 'Extract', schema: str = 'Extract', minimise=False):
        if minimise:
            raise NotImplementedError('Cannot minimise data when using streaming')
        self.stream_sql_to_hyper(
            file_path=file_path,
            table=table,
            schema=schema
        )

    def get_total(self, measure=None):
        """
        For totals when streaming data we need to run a totals SQL query separate
        from the main query and use the results of this
        :return: the total value of the query
        """
        logger.info("Building totals query")
        measure = self.__get_measure__(measure)
        if self.configuration.query_builder is not None:
            from mario.query_builder import QueryBuilder
            query_builder: QueryBuilder = self.configuration.query_builder(
                configuration=self.configuration,
                metadata=self.metadata,
                dataset_specification=self.dataset_specification)
            totals_query = query_builder.create_totals_query(measure=measure)
        else:
            raise NotImplementedError

        totals_df = pd.read_sql(totals_query[0], self.get_connection(), params=totals_query[1])
        return totals_df.iat[0, 0]

    def stream_sql_to_hyper(self,
                            file_path: str,
                            table: str = 'Extract',
                            schema: str = 'Extract',
                            validate: bool = False,
                            allow_nulls: bool = True,
                            minimise: bool = False,
                            chunk_size: int = 100000):
        """
        Write From SQL to .hyper using streaming. No data is held in memory
        apart from chunks of rows as they are read.
        Optionally, data can be validated as it is read.
        """
        self.__build_query__()
        logger.info("Executing query")
        from tableauhyperapi import TableName
        from pantab import frame_to_hyper

        connection = self.get_connection()
        table_name = TableName(schema, table)
        for df in pd.read_sql(self._query[0], connection, chunksize=chunk_size):
            if validate or minimise:
                self._data = df
                if validate:
                    self.validate_data(allow_nulls=allow_nulls)
                if minimise:
                    self.__minimise_data__()
                    df = self._data
            frame_to_hyper(df, database=file_path, table=table_name, table_mode='a')

    def stream_sql_to_csv(self,
                          file_path,
                          validate: bool = False,
                          allow_nulls: bool = True,
                          chunk_size: int = 100000,
                          compress_using_gzip: bool = False,
                          minimise: bool = False
                          ):
        """
        Write From SQL to CSV using streaming. No data is held in memory
        apart from chunks of rows as they are read.
        Optionally, data can be validated as it is read.
        """
        self.__build_query__()
        logger.info("Executing query")
        connection = self.get_connection()

        if compress_using_gzip:
            compression_options = dict(method='gzip')
            file_path = file_path + '.gz'
        else:
            compression_options = None

        mode = 'w'
        header = True
        for df in pd.read_sql(self._query[0], connection, chunksize=chunk_size):
            if validate or minimise:
                self._data = df
                if validate:
                    self.validate_data(allow_nulls=allow_nulls)
                if minimise:
                    self.__minimise_data__()
                    df = self._data
            df.to_csv(file_path, mode=mode, header=header, index=False, compression=compression_options)
            if header:
                header = False
                mode = "a"

        return file_path

    def stream_sql_to_csv_using_bcp(self,
                                    table_name: str,
                                    output_file_path: str,
                                    server_url: str,
                                    database_name: str,
                                    use_view=False,
                                    keep_existing_table=False,
                                    delete_when_finished=False):
        """
        Builds table on server, then extracts it to CSV using the bcp program. No data is held in
        memory. Data cannot be validated without reading it in from the generated CSV.
        TODO Note that the generated SQL must start with 'SELECT * FROM' - make this more generic
        TODO Provide an option to not drop/create table if it already exists.
        """
        self.__build_query__()
        logger.info("Executing query")
        from sqlalchemy import text
        import subprocess

        # Get a connection
        connection = self.get_connection()

        table_or_view = 'TABLE' if not use_view else 'VIEW'

        # Drop the table if it already exists
        drop_table_sql = f'DROP {table_or_view} IF EXISTS {self.configuration.schema}.{table_name};'
        connection.execute(text(drop_table_sql))
        connection.commit()
        logger.debug(f'drop {table_or_view}: {drop_table_sql}')

        # Create a table or view using SQL
        sql = self._query[0]
        if use_view:
            create_table_sql = f'CREATE VIEW dbo.{table_name} AS ' + sql
        else:
            create_table_sql = sql.replace('SELECT * FROM (', f'SELECT * INTO dbo.{table_name} FROM (')
        connection.execute(text(create_table_sql))
        connection.commit()
        logger.debug(f'create {table_or_view}: {create_table_sql}')

        # SQL to obtain the column names from the table
        bcp_columns_sql = f'SELECT COLUMN_NAME FROM INFORMATION_SCHEMA.COLUMNS WHERE TABLE_NAME = \'{table_name}\''
        bcp_columns = [r[0] for r in connection.execute(text(bcp_columns_sql)).fetchall()]
        bcp_columns_literal = ', '.join(["'" + a + "'" for a in bcp_columns])
        bcp_columns_select = ["CHAR(34)+CAST([" + a + "] AS VARCHAR(max))+CHAR(34)" for a in bcp_columns]

        # Union the column names with the extract to get the complete query
        bcp_select = f"SELECT {bcp_columns_literal} UNION ALL SELECT {','.join(bcp_columns_select)} FROM [dbo].{table_name}"

        # Export to CSV
        user = self.configuration.user
        bcp_code = f'bcp "{bcp_select}" queryout {output_file_path} -d {database_name} -c -C RAW -U {user} -S {server_url} -G'  # 65001 for utf-8 // RAW for same data
        logger.debug('bcp command: ' + bcp_code)
        subprocess.call(bcp_code)

        # Delete table/view when finished
        if delete_when_finished:
            connection.execute(text(drop_table_sql))
            connection.commit()

        # Close connection when done
        connection.close()


class DataFrameExtractor(DataExtractor):
    """
    Wrapper for a DataExtractor using an existing DataFrame
    """

    def __init__(self,
                 dataset_specification: DatasetSpecification,
                 metadata: Metadata,
                 dataframe: DataFrame
                 ):
        super().__init__(configuration=Configuration(), dataset_specification=dataset_specification, metadata=metadata)
        self._data = dataframe


class PartitioningExtractor(DataExtractor):
    """
    A data extractor that loads from SQL in batches using a specified constraint
    to partition by
    """
    def __init__(self,
                 configuration: Configuration,
                 dataset_specification: DatasetSpecification,
                 metadata: Metadata,
                 partition_column: str
                 ):
        super().__init__(configuration, dataset_specification, metadata)
        self._data = None
        self.partition_column = partition_column

    def get_connection(self):
        from sqlalchemy import create_engine

        if self.configuration.hook is not None:
            return self.configuration.hook.get_conn()

        engine = create_engine(self.configuration.connection_string)
        connection = engine.connect().execution_options(stream_results=True)
        return connection

    def __get_partition_values__(self):
        for constraint in self.dataset_specification.constraints:
            if constraint.item == self.partition_column:
                return constraint.allowed_values

    def __build_query_using_partition__(self, partition_value):
        logger.info(f"Building query for partition {self.partition_column} with value {partition_value}")
        from copy import deepcopy
        specification = deepcopy(self.dataset_specification)

        partition_found = False
        for constraint in specification.constraints:
            if constraint.item == self.partition_column:
                constraint.allowed_values = [partition_value]
                partition_found = True

        if not partition_found:
            raise ValueError("Partition not found in constraints")

        if self.configuration.query_builder is not None:
            from mario.query_builder import QueryBuilder
            query_builder: QueryBuilder = self.configuration.query_builder(
                configuration=self.configuration,
                metadata=self.metadata,
                dataset_specification=specification)
            query = query_builder.create_query()
        else:
            raise NotImplementedError

        return query

    def __load_from_sql_using_partition__(self, partition_value):
        logger.info("Executing query")

        query = self.__build_query_using_partition__(partition_value=partition_value)
        from sqlalchemy import create_engine
        engine = create_engine(self.configuration.connection_string)
        if self._data is None:
            self._data = pd.read_sql(sql=query[0], con=engine.connect(), params=query[1])
        else:
            df = pd.read_sql(sql=query[0], con=engine.connect(), params=query[1])
            self._data = pd.concat([self._data, df], axis=0)

    def __load_from_sql__(self):
        for value in self.__get_partition_values__():
            self.__load_from_sql_using_partition__(partition_value=value)

    def get_data_frame(self, minimise=True) -> DataFrame:
        raise NotImplementedError()

    def get_total(self, measure=None):
        raise NotImplementedError()

    def save_data_as_hyper(self,
                           file_path: str,
                           table: str = 'Extract',
                           schema: str = 'Extract',
                           minimise=True):
        self.stream_sql_to_hyper(file_path=file_path, table=table, schema=schema, minimise=minimise)

    def save_data_as_csv(self, file_path: str, minimise=True, compress_using_gzip=False):
        self.stream_sql_to_csv(
            file_path=file_path,
            minimise=minimise,
            compress_using_gzip=compress_using_gzip
        )

    def stream_sql_to_csv(self,
                          file_path,
                          validate: bool = False,
                          allow_nulls: bool = True,
                          chunk_size: int = 100000,
                          compress_using_gzip: bool = False,
                          minimise: bool = False
                          ):
        """
        Write From SQL to CSV using streaming. No data is held in memory
        apart from chunks of rows as they are read.
        Optionally, data can be validated as it is read.
        """
        from mario.query_builder import get_formatted_query
        logger.info("Executing query")
        connection = self.get_connection()

        if compress_using_gzip:
            compression_options = dict(method='gzip')
            file_path = file_path + '.gz'
        else:
            compression_options = None

        mode = 'w'
        header = True
        for partition_value in self.__get_partition_values__():
            query = self.__build_query_using_partition__(partition_value=partition_value)
            for df in pd.read_sql(get_formatted_query(query[0],query[1]), connection, chunksize=chunk_size):
                if validate or minimise:
                    self._data = df
                    if validate:
                        self.validate_data(allow_nulls=allow_nulls)
                    if minimise:
                        self.__minimise_data__()
                        df = self._data
                df.to_csv(file_path, mode=mode, header=header, index=False, compression=compression_options)
                if header:
                    header = False
                    mode = "a"

        return file_path

    def stream_sql_to_hyper(self,
                            file_path: str,
                            table: str = 'Extract',
                            schema: str = 'Extract',
                            validate: bool = False,
                            allow_nulls: bool = True,
                            minimise: bool = False,
                            chunk_size: int = 100000):
        """
        Write From SQL to .hyper using streaming. No data is held in memory
        apart from chunks of rows as they are read.
        Optionally, data can be validated as it is read.
        """
        self.__build_query__()
        logger.info("Executing query")
        from tableauhyperapi import TableName
        from pantab import frame_to_hyper
        from mario.query_builder import get_formatted_query

        connection = self.get_connection()
        table_name = TableName(schema, table)
        for partition_value in self.__get_partition_values__():
            query = self.__build_query_using_partition__(partition_value=partition_value)
            for df in pd.read_sql(get_formatted_query(query[0], query[1]), connection, chunksize=chunk_size):
                if validate or minimise:
                    self._data = df
                    if validate:
                        self.validate_data(allow_nulls=allow_nulls)
                    if minimise:
                        self.__minimise_data__()
                        df = self._data
                logger.info(f"Saving {chunk_size} rows to file")
                frame_to_hyper(df, database=file_path, table=table_name, table_mode='a')





import logging
from enum import Enum

from mario.data_extractor import Configuration
from mario.dataset_specification import DatasetSpecification
from mario.metadata import Metadata, Item

logger = logging.getLogger(__name__)


class DataTypes(Enum):
    """
    Base data types, that are mapped onto specific
    types used in Pandas, Hyper, SQL etc
    """
    TEXT = 'text'
    DATE = 'date'
    INT = 'int'
    DOUBLE = 'double'
    DATETIME = 'datetime'
    OBJECT = 'object'


class Validator:
    """
    Base class for validators. Extended for specific data implementations
    such as Pandas Dataframe, Tableau HyperFile etc.
    """

    def __init__(self,
                 dataset_specification: DatasetSpecification,
                 metadata: Metadata
                 ):
        self.dataset_specification = dataset_specification
        self.metadata = metadata
        self.errors = []
        self.warnings = []

    def __get_column_name__(self, item: Item):
        """ Returns the column name for a metadata item"""
        if item.get_property('output_name') is not None:
            return item.get_property('output_name')
        elif item.get_property('physical_column_name') is not None:
            return item.get_property('physical_column_name')
        else:
            return item.name

    def __get_column_data_type__(self, item: Item):
        """ Returns the type of a metadata item from the data"""
        raise NotImplementedError()

    def __get_column_values__(self, item: Item):
        """ Returns an array of the unique values in the data for an item """
        raise NotImplementedError()

    def __get_minimum_maximum_values__(self, item:Item):
        """ Returns the minimum value of an item from the data """
        raise NotImplementedError()

    def __contains_nulls__(self, item: Item):
        """ Returns True if the item has any Null values in the data """
        return None in self.__get_column_values__(item)

    def __get_data_for_hierarchy__(self, name):
        """ Returns the dataframe for a hierarchy """
        raise NotImplementedError()

    def __get_column_with_segmentation__(self, item:Item, segmentation: str):
        """ Returns the data series for an item grouped by the segmentation column specified """
        raise NotImplementedError()

    def __check_hierarchy__(self, name):
        """ Checks whether the named hierarchy conforms to a tree structure """
        # Create a dictionary to store the hierarchy
        hierarchy = {}

        # Get the levels
        levels = self.metadata.get_hierarchy(name)

        # Get the data
        df = self.__get_data_for_hierarchy__(name)

        # Group by all levels except the last one
        grouped = df.groupby(levels[:-1])[levels[-1]].apply(list).reset_index()

        # Check for duplicates at the last level
        for _, row in grouped.iterrows():
            higher_levels = tuple(row[levels[:-1]])
            values = row[levels[-1]]
            for value in values:
                if value not in hierarchy:
                    hierarchy[value] = higher_levels
                elif hierarchy[value] != higher_levels:
                    self.errors.append(f"Inconsistent hierarchy: {value} at level {levels[-1]} is represented in multiple higher level categories {hierarchy[value]} and {higher_levels}.")

    def __check_for_anomalies__(self, series, threshold=1.75):
        """
        Checks a series for anomalies by looking for absolute z-scores above threshold
        :param series: the series to check
        :param threshold:  the threshold level
        :return: True if an anomaly is present
        """
        import numpy as np
        mean = series.mean()
        std = series.std()
        z_scores = (series - mean) / std
        return np.any(np.abs(z_scores) > threshold)

    def check_hierarchies(self):
        """
        Checks the validity of hierarchies, i.e. that they form a tree structure and
        adds any anomalies found to errors/warnings
        """
        for hierarchy in self.metadata.get_hierarchies():
            self.__check_hierarchy__(hierarchy)

    def check_item_for_anomalies(self, item: str, segmentation: str):
        item = self.metadata.get_metadata(item)
        subset = self.__get_column_with_segmentation__(item, segmentation)
        if self.__check_for_anomalies__(subset):
            self.warnings.append(
                f"Validation warning: '{item.name}' has potentially anomalous data when segmented by '{segmentation}'")

    def check_category_anomalies(self, segmentation: str):
        """
        Checks to see if we get anomalies in how categories are split e.g. over time
        and adds them to warnings
        :param segmentation: the field to segment by
        :return: None
        """
        for dimension in self.dataset_specification.dimensions:
            self.check_item_for_anomalies(dimension, segmentation)

    def check_data_type(self, item: Item):
        """ Checks whether the item has the correct data type """
        expected_data_type = item.get_property('datatype')
        if expected_data_type is None:
            return True

        expected_data_type = DataTypes[expected_data_type.upper()]
        actual_data_type = self.__get_column_data_type__(item)

        if actual_data_type == expected_data_type:
            return True

        if actual_data_type == DataTypes.DATETIME and expected_data_type == DataTypes.DATE:
            self.warnings.append(f"Validation warning: '{item.name}': expected {expected_data_type} but was {actual_data_type}")
            return False

        if actual_data_type == DataTypes.OBJECT and expected_data_type == DataTypes.DATE:
            self.warnings.append(f"Validation warning: '{item.name}': expected {expected_data_type} but was {actual_data_type}"
                                 f". This is probably due to a limitation of Pandas not having a native 'Date' type")
            return False

        self.errors.append(f"Validation error: '{item.name}' is of the wrong type - expected {expected_data_type}, actual {actual_data_type}")
        return False

    def check_column_present(self, item: Item):
        """ Checks whether the column for an item is present in the data """
        raise NotImplementedError()

    def check_nulls(self, item: Item, allow_nulls=True):
        """ Checks if there are any null values for an item """
        if self.__contains_nulls__(item):
            if allow_nulls:
                self.warnings.append(f"Validation warning: '{item.name}' contains NULLs")
            else:
                self.errors.append(f"Validation error: '{item.name}' contains NULLs")

    def check_domain(self, item: Item):
        """ Checks that the values for an item conform to the domain in its specification """
        if item.get_property('domain') is not None:
            data_domain = self.__get_column_values__(item)
            metadata_domain = item.get_property('domain')
            for element in data_domain:
                if element not in metadata_domain:
                    self.errors.append(f"Validation error: '{str(element)}' is not in domain of '{item.name}'")
            for metadata_element in metadata_domain:
                if metadata_element not in data_domain:
                    self.warnings.append(f"Validation warning: '{str(metadata_element)}' is in domain of '{item.name}' but not present in the data")

    def check_range(self, item: Item):
        """ Checks whether the values of an item fall within expected range """
        if item.get_property('range') is not None:
            min_value = item.get_property('range')[0]
            max_value = item.get_property('range')[1]
            data_min, data_max = self.__get_minimum_maximum_values__(item)
            if data_min < min_value:
                self.errors.append(f"Validation error: '{item.name}': '{str(data_min)}' is less than '{str(min_value)}'")
            if data_max > max_value:
                self.errors.append(f"Validation error: '{item.name}': '{str(data_max)}' is greater than '{str(max_value)}'")

    def check_pattern_match(self, item: Item):
        """ Checks that the values for an item match the pattern in the specification """
        import re
        if item.get_property('pattern') is not None:
            values = self.__get_column_values__(item)
            pattern = re.compile(item.get_property('pattern'))
            for value in values:
                if value is not None:
                    if not re.match(pattern, str(value)):
                        self.errors.append(f"Validation error: '{item.name}': '{str(value)}' does not match the pattern '{str(item.get_property('pattern'))}'")

    def check_quality_checks(self, item: Item):
        """ Checks whether an item has any quality rules used in validation """
        if item.get_property('formula') is None:
            if item.get_property('range') is None:
                if item.get_property('domain') is None:
                    if item.get_property('pattern') is None:
                        self.warnings.append(f"Validation warning: '{item.name}' has no quality rules.")

    def validate_data_item(self, item, allow_nulls=True):
        metadata = self.metadata.get_metadata(item)
        if self.check_column_present(metadata):
            self.check_nulls(metadata, allow_nulls)
            self.check_domain(metadata)
            self.check_range(metadata)
            self.check_quality_checks(metadata)
            self.check_data_type(metadata)
            self.check_pattern_match(metadata)

    def validate_data(self, allow_nulls=True, check_hierarchies=False, detect_anomalies=False, segmentation=None):
        """
        Performs validation of data
        :param allow_nulls: If True, nulls are shown as warnings rather than errors
        :param check_hierarchies: If True, perform hierarchy consistency check
        :param detect_anomalies: If True, perform anomaly detection
        :param segmentation: Optional - field name to use for anomaly detection
        :return: True if no validation errors are found
        """
        for item in self.dataset_specification.items:
            self.validate_data_item(item, allow_nulls)
        if check_hierarchies:
            self.check_hierarchies()
        if detect_anomalies:
            if segmentation is not None:
                self.check_category_anomalies(segmentation)
            else:
                raise ValueError("A segmentation column, e.g. year, must be supplied to detect anomalies.")

        for warning in self.warnings:
            logger.warning(warning)
        for validation_error in self.errors:
            logger.error(validation_error)
        if len(self.errors) != 0:
            raise ValueError("Data validation failed - check the logs for details")
        return True


class DataFrameValidator(Validator):
    """
    A validator for Pandas Dataframes, or data extractors that
    use a dataframe as their internal representation.
    """
    def __init__(self,
                 dataset_specification: DatasetSpecification,
                 metadata: Metadata,
                 data,
                 ):
        super().__init__(dataset_specification, metadata)
        self.data = data

    def __get_column_data_type__(self, item: Item):
        column = self.__get_column_name__(item)
        data_type = str(self.data[column].dtype).lower()
        if data_type in ['category', 'string']:
            return DataTypes.TEXT
        if data_type in ['float64']:
            return DataTypes.DOUBLE
        if data_type in ['int', 'int64']:
            return DataTypes.INT
        if data_type in ['datetime64', 'datetime64[ns]']:
            return DataTypes.DATETIME
        if data_type in ['object']:
            return DataTypes.OBJECT
        return data_type

    def __get_column_values__(self, item: Item):
        import pandas as pd
        column = self.__get_column_name__(item)
        values = self.data[column].replace({pd.NA: None}).unique()
        return list(values)

    def __get_minimum_maximum_values__(self, item:Item):
        column = self.__get_column_name__(item)
        data_min = self.data[column].min()
        data_max = self.data[column].max()
        return data_min, data_max

    def __get_data_for_hierarchy__(self, name):
        fields = self.metadata.get_hierarchy(name)
        return self.data[fields].drop_duplicates().reset_index(drop=True)

    def check_column_present(self, item: Item):
        if item.get_property('formula') is None:
            column = self.__get_column_name__(item)
            if column not in self.data.columns:
                self.errors.append(f"Validation error: '{item.name}' in specification is missing from dataset")
                return False
            return True
        return False

    def __get_column_with_segmentation__(self, item:Item, segmentation: str):
        column = self.__get_column_name__(item)
        return self.data.groupby(segmentation)[column].nunique()


class HyperValidator(Validator):
    """
    A validator for hyper files
    """

    def __init__(self,
                 dataset_specification: DatasetSpecification,
                 metadata: Metadata,
                 hyper_file_path
                 ):
        super().__init__(dataset_specification, metadata)
        from mario.hyper_utils import get_default_table_and_schema
        self.hyper_file_path = hyper_file_path
        schema, table = get_default_table_and_schema(self.hyper_file_path)
        self.table = table
        self.schema = schema

    def __get_data_for_hierarchy__(self, name):
        from pantab import frame_from_hyper_query
        fields = ', '.join(f'"{s}"' for s in self.metadata.get_hierarchy(name))
        query = f"""
                     SELECT
                         {fields} 
                     FROM "{self.schema}"."{self.table}" 
                     GROUP BY {fields} 
        """
        from tableauhyperapi import HyperProcess, Telemetry, Connection
        with HyperProcess(Telemetry.DO_NOT_SEND_USAGE_DATA_TO_TABLEAU, 'test') as hyper:
            with Connection(hyper.endpoint, self.hyper_file_path) as connection:
                results_df = frame_from_hyper_query(connection,query)
        return results_df

    def __get_column_with_segmentation__(self, item:Item, segmentation: str):
        from pantab import frame_from_hyper_query
        column = self.__get_column_name__(item)
        query = f"""
                    SELECT COUNT(DISTINCT "{column}")
                    FROM "{self.schema}"."{self.table}" 
                    GROUP BY "{segmentation}"
        """
        from tableauhyperapi import HyperProcess, Telemetry, Connection
        with HyperProcess(Telemetry.DO_NOT_SEND_USAGE_DATA_TO_TABLEAU, 'test') as hyper:
            with Connection(hyper.endpoint, self.hyper_file_path) as connection:
                results_df = frame_from_hyper_query(connection,query)
        return results_df

    def check_column_present(self, item: Item):
        column = self.__get_column_name__(item)
        from mario.hyper_utils import check_column_exists
        if item.get_property('formula') is None:
            if not check_column_exists(
                    hyper_file_path=self.hyper_file_path,
                    column=column,
                    table=self.table,
                    schema=self.schema
            ):
                self.errors.append(f"Validation error: '{item.name}' in specification is missing from dataset")
                return False
            return True
        return False

    def __get_column_data_type__(self, item: Item):
        from mario.hyper_utils import get_table
        from tableauhyperapi import TypeTag
        column = self.__get_column_name__(item)
        table = get_table(hyper_path=self.hyper_file_path, table=self.table, schema=self.schema)
        datatype = table.get_column_by_name(column).type.tag
        if datatype in [TypeTag.TEXT, TypeTag.CHAR]:
            return DataTypes.TEXT
        if datatype in [TypeTag.BIG_INT, TypeTag.INT, TypeTag.SMALL_INT]:
            return DataTypes.INT
        if datatype in [TypeTag.DOUBLE, TypeTag.NUMERIC]:
            return DataTypes.DOUBLE
        if datatype in [TypeTag.DATE]:
            return DataTypes.DATE
        if datatype in [TypeTag.TIMESTAMP, TypeTag.TIMESTAMP_TZ]:
            return DataTypes.DATETIME
        return str(datatype)

    def __get_minimum_maximum_values__(self, item):
        from tableauhyperapi import HyperProcess, Telemetry, Connection
        column = self.__get_column_name__(item)
        with HyperProcess(Telemetry.DO_NOT_SEND_USAGE_DATA_TO_TABLEAU, 'test') as hyper:
            with Connection(hyper.endpoint, self.hyper_file_path) as connection:
                with connection.execute_query(
                        'SELECT MIN("' + column + '") FROM "' + self.schema + '"."' + self.table + '"') as result:
                    min_data = [item for row in list(result) for item in row][0]
                with connection.execute_query(
                        'SELECT MAX("' + column + '") FROM "' + self.schema + '"."' + self.table + '"') as result:
                    max_data = [item for row in list(result) for item in row][0]
        return min_data, max_data

    def __get_column_values__(self, item: Item):
        from tableauhyperapi import HyperProcess, Telemetry, Connection
        column = self.__get_column_name__(item)
        with HyperProcess(Telemetry.DO_NOT_SEND_USAGE_DATA_TO_TABLEAU, 'test') as hyper:
            with Connection(hyper.endpoint, self.hyper_file_path) as connection:
                with connection.execute_query(
                        'SELECT DISTINCT "' + column + '" FROM "' + self.schema + '"."' + self.table + '"') as result:
                    rows = list(result)
                    hyper_domain = [item for row in rows for item in row]
        return hyper_domain


class SqlValidator(Validator):
    def __init__(self,
                 dataset_specification: DatasetSpecification,
                 metadata: Metadata,
                 configuration: Configuration
                 ):
        super().__init__(dataset_specification, metadata)
        self.connection = self.__get_connection__(configuration.connection_string)
        self.schema = configuration.schema
        self.view = configuration.view

    def __get_connection__(self, connection_string: str):
        from sqlalchemy import create_engine
        engine = create_engine(connection_string)
        connection = engine.connect().execution_options(stream_results=True)
        return connection

    def __get_column_with_segmentation__(self, item: Item, segmentation: str):
        import pandas as pd
        column = self.__get_column_name__(item)
        sql = f"""
                    SELECT COUNT(DISTINCT "{column}")
                    FROM "{self.schema}"."{self.view}" 
                    GROUP BY "{segmentation}"
        """
        return pd.read_sql(sql, self.connection)

    def __get_column_data_type__(self, item: Item):
        import pandas as pd
        column = self.__get_column_name__(item)
        sql = f"SELECT DATA_TYPE AS dt FROM INFORMATION_SCHEMA.COLUMNS WHERE TABLE_NAME='{self.view}'" \
              f" AND TABLE_SCHEMA='{self.schema}'" \
              f" AND COLUMN_NAME='{column}'"
        df = pd.read_sql(sql, self.connection)
        type = df.at[0, 'dt']
        if type in ['tinyint', 'int']:
            return DataTypes.INT
        elif type in ['char', 'varchar', 'nvarchar', 'varbinary', 'text']:
            return DataTypes.TEXT
        elif type in ['real', 'decimal', 'double precision']:
            return DataTypes.DOUBLE
        elif type in ['date']:
            return DataTypes.DATE
        elif type in ['datetime']:
            return DataTypes.DATETIME
        else:
            return type

    def __get_minimum_maximum_values__(self, item:Item):
        import pandas as pd
        column = self.__get_column_name__(item)
        sql = f"SELECT min(\"{column}\") AS min_value, max(\"{column}\") as max_value FROM {self.schema}.{self.view}"
        df = pd.read_sql(sql, self.connection)
        min_value = df.at[0, 'min_value']
        max_value = df.at[0, 'max_value']
        return min_value, max_value

    def __get_column_values__(self, item: Item):
        import pandas as pd
        column = self.__get_column_name__(item)
        sql = f"SELECT DISTINCT \"{column}\" AS checkfield FROM {self.schema}.{self.view}"
        df = pd.read_sql(sql, self.connection)
        values = df['checkfield'].to_list()
        return values

    def __get_data_for_hierarchy__(self, name):
        import pandas as pd
        fields = ', '.join(f'"{s}"' for s in self.metadata.get_hierarchy(name))
        sql = f"""
                     SELECT
                         {fields} 
                     FROM "{self.schema}"."{self.view}" 
                     GROUP BY {fields} 
        """
        return pd.read_sql(sql, self.connection)

    def check_column_present(self, item: Item):
        import pandas as pd
        if item.get_property('formula') is not None:
            return False
        sql = f"SELECT COLUMN_NAME as col FROM INFORMATION_SCHEMA.COLUMNS WHERE TABLE_NAME='{self.view}' AND TABLE_SCHEMA='{self.schema}'"
        df = pd.read_sql(sql, self.connection)
        values = df['col'].to_list()
        if item.name not in values:
            self.errors.append(f"Validation error: '{item.name}' in specification is missing from dataset")
            return False
        return True



import os

from mario.dataset_specification import dataset_from_json, dataset_from_manifest, dataset_from_excel


def test_import_dataset():
    specification_file = os.path.join('test', 'dataset.json')
    specification = dataset_from_json(file_path=specification_file)

    assert specification.collection == 'superstore'
    assert len(specification.measures) == 4
    assert specification.get_property('description') is not None


def test_import_dataset_from_manifest():
    specification_file = os.path.join('test', 'manifest.json')
    specification = dataset_from_manifest(file_path=specification_file)

    assert specification.name == 'test@jisc.ac.uk'
    assert specification.collection == 'manifest_test'


def test_import_dataset_from_excel():
    specification_file = os.path.join('test', 'SpecificationInputTemplate.xlsx')
    specification = dataset_from_excel(file_path=specification_file)

    assert specification.name == 'Item 1'
    assert specification.get_property('Item Number') == 'Item 1'
    assert specification.collection == 'Student Record and AP Student Record'
    assert specification.measures == ['FPE']
    assert specification.constraints[0].allowed_values == [2021, 2022, 2023]
    assert specification.constraints[1].item == 'Onward use category'
    assert specification.constraints[1].allowed_values == [1]
    assert specification.dimensions == ['Sex', 'HE Provider']


import os

from mario.data_extractor import Configuration, DataExtractor
from mario.dataset_specification import dataset_from_json, dataset_from_manifest
from mario.query_builder import SubsetQueryBuilder, ViewBasedQueryBuilder
from mario.metadata import metadata_from_json, metadata_from_manifest


def test_manifest_subset():
    dataset = dataset_from_manifest(os.path.join('test', 'manifest.json'))
    metadata = metadata_from_manifest(os.path.join('test', 'manifest.json'))
    configuration = Configuration(
        view='v_student_fpe',
        schema='dbo',
        query_builder=SubsetQueryBuilder
    )
    extractor = DataExtractor(configuration=configuration, dataset_specification=dataset, metadata=metadata)
    path = os.path.join('output', dataset.collection, dataset.name + '.sql')
    os.makedirs(os.path.join('output', dataset.collection), exist_ok=True)
    extractor.save_query(file_path=path)


def test_spec_subset():
    dataset = dataset_from_json(os.path.join('test', 'dataset.json'))
    metadata = metadata_from_json(os.path.join('test', 'metadata.json'))
    configuration = Configuration(
        view='v_student_fpe',
        schema='dbo',
        query_builder=SubsetQueryBuilder
    )
    extractor = DataExtractor(configuration=configuration, dataset_specification=dataset, metadata=metadata)
    path = os.path.join('output', dataset.collection, dataset.name + '_subset.sql')
    os.makedirs(os.path.join('output', dataset.collection), exist_ok=True)
    extractor.save_query(file_path=path)


def test_spec_view():
    dataset = dataset_from_json(os.path.join('test', 'dataset.json'))
    metadata = metadata_from_json(os.path.join('test', 'metadata.json'))
    configuration = Configuration(
        view='v_student_fpe',
        schema='dbo',
        query_builder=ViewBasedQueryBuilder
    )
    extractor = DataExtractor(configuration=configuration, dataset_specification=dataset, metadata=metadata)
    path = os.path.join('output', dataset.collection, dataset.name + '_view.sql')
    os.makedirs(os.path.join('output', dataset.collection), exist_ok=True)
    extractor.save_query(file_path=path)


import json
from typing import List
from mario.base import MarioBase


class Constraint:
    def __init__(self):
        self.item = None
        self.allowed_values = []


class DatasetSpecification(MarioBase):
    """
    Base class/interface for dataset specifications. Use an implementation specific to the
    format of specification used.
    """

    def __init__(self):
        super().__init__()
        self._name = ''
        self._collection: str = ''
        self._measures: List[str] = []
        self._dimensions: List[str] = []
        self._constraints: List[Constraint] = []

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        self._name = value

    @property
    def collection(self):
        return self._collection

    @collection.setter
    def collection(self, value):
        self._collection = value

    @property
    def enquiry(self):
        return ''

    @property
    def measures(self):
        return self._measures

    @measures.setter
    def measures(self, value):
        self._measures = value

    @property
    def dimensions(self):
        return self._dimensions

    @dimensions.setter
    def dimensions(self, value):
        self._dimensions = value

    @property
    def constraints(self):
        return self._constraints

    @constraints.setter
    def constraints(self, value):
        self._constraints = value

    @property
    def items(self):
        return self.dimensions + self.measures

    def save(self, file_path: str):
        json_representation = {
            "name": self.name,
            "collection": self.collection,
            "dimensions": self.dimensions,
            "measures": self.measures
        }
        with open(file_path, 'w', encoding='utf-8') as file:
            json.dump(json_representation, file, default=vars)


def dataset_from_json(file_path: str = None) -> DatasetSpecification:
    with open(file_path, mode='r', encoding='utf-8') as source_file:
        spec = json.load(source_file)

    dataset_specification = DatasetSpecification()
    dataset_specification.name = spec['name']
    dataset_specification.collection = spec['collection']
    dataset_specification.measures = spec['measures']
    dataset_specification.dimensions = spec['dimensions']
    dataset_specification.add_properties(source=spec, exclude=['name', 'collection', 'measures', 'dimensions'])

    return dataset_specification


def dataset_from_manifest(file_path: str = None):
    with open(file_path, mode='r', encoding='utf-8') as source_file:
        spec = json.load(source_file)

    dataset_specification = DatasetSpecification()

    fields = []
    constraints = []
    for dimension in spec['items']:
        if 'allowedValues' in dimension and len(dimension['allowedValues']) > 0:
            constraint = Constraint()
            constraint.item = dimension['fieldName']
            constraint.allowed_values = dimension['allowedValues']
            constraints.append(constraint)
        else:
            fields.append(dimension['fieldName'])
    if 'onwardUseCategory' in spec:
        constraint = Constraint()
        constraint.item = 'Onward use category '+str(spec['onwardUseCategory'])
        constraint.allowed_values = [1]
        constraints.append(constraint)
    if 'years' in spec:
        constraint = Constraint()
        constraint.item = 'Academic year start'
        constraint.allowed_values = get_year_starts(spec['years'])
        constraints.append(constraint)

    dataset_specification.constraints = constraints
    dataset_specification.measures = [spec['measure']]
    dataset_specification.dimensions = fields
    dataset_specification.name = spec['client']
    dataset_specification.collection = spec['orderRef']
    dataset_specification.add_properties(source=spec, exclude=['name', 'orderRef', 'measure', 'items'])

    return dataset_specification


def get_year_starts(years):
    """
    Converts the array of academic year strings for a request into an array of year start integers
    :return: an array of integers for each year start
    """
    year_starts = []
    for year in years:
        year_start = int(year.split('/')[0])
        year_starts.append(year_start)
    return year_starts


def dataset_from_excel(file_path: str):
    from openpyxl import load_workbook
    from openpyxl.worksheet.worksheet import Worksheet

    content = {
        'Enquiry Number': 'A3',
        'Item Number': 'C2',
        'Organisation Name': 'A6',
        'Licence expiry date': 'A9',
        'Data Format': 'C5',
        'Onward use': 'C13',
        'Additional restrictions': 'C15'
    }

    dataset_specification = DatasetSpecification()

    # Load sheet
    workbook = load_workbook(filename=file_path, read_only=True)
    sheet: Worksheet = workbook.get_sheet_by_name('InputTemplate')

    # Read general spec values
    dataset_specification.name = sheet['C2'].value
    dataset_specification.collection = sheet['C8'].value
    for prop in content.items():
        value = sheet[prop[1]].value
        dataset_specification.set_property(prop[0], value)

    # Read fields
    cell_range = 'E3:E100'
    for row in sheet[cell_range]:
        for cell in row:
            if cell.value:
                dataset_specification.dimensions.append(cell.value)

    # Measure
    dataset_specification.measures = [sheet['C22'].value]

    # Constraints
    constraints = []

    # Years
    spec_year = sheet['C11'].value
    year_start = int(spec_year[0:4])
    year_end = year_start
    spec_year = spec_year.replace("—", "-")
    if spec_year.find('-') != -1:
        year_end = int(spec_year.split('-')[1].strip())

    years_constraint = Constraint()
    years_constraint.item = 'Academic year'
    years_constraint.allowed_values = list(range(year_start, year_end + 1))
    constraints.append(years_constraint)

    # Onward use
    onward_use_constraint = Constraint()
    onward_use_constraint.item = 'Onward use category'
    onward_use_constraint.allowed_values = [sheet['C13'].value]
    constraints.append(onward_use_constraint)

    # Other restrictions - c15 to c19
    for cell in ['C15', 'C16', 'C17', 'C18', 'C19']:
        if sheet[cell].value is not None:
            constraint = Constraint()
            constraint.item = sheet[cell].value
            constraints.append(constraint)

    dataset_specification.constraints = constraints

    # Deal with windows greedy file handler madness
    workbook.close()
    import gc
    gc.collect()

    return dataset_specification



from typing import List
from openpyxl import Workbook
import pandas as pd


def replace_pivot_cache_with_subset(worksheet, field, value) -> None:
    """
    Replaces the cached data for a pivot with a subset filtered by
    the supplied field and value
    :param worksheet: the sheet to filter
    :param field: the field name
    :param value: the field value to filter by
    :return: None
    """
    pivot = worksheet._pivots[0]
    cached_data = pivot.cache
    subset = get_subset(cached_data, field, value)
    pivot_cache = worksheet._pivots[0].cache
    pivot_cache.records.r = subset


def get_subset(pivot_cache, field_name, field_value):
    """
    Returns a subset from pivot cache data using the field
    name and field value as a filter
    :param pivot_cache: the cache to filter
    :param field_name: the name of the field to filter by
    :param field_value: the value to filter by
    :return: a filtered pivot cache
    """
    # Get the index of the field name
    field_index = None
    cache_field = None
    field_value_index = None

    for idx, field in enumerate(pivot_cache.cacheFields):
        if field.name == field_name:
            cache_field = field
            field_index = idx
            break

    # Get the index of the field value
    # Map the indices to actual values
    field_items = cache_field.sharedItems._fields
    for index, item in enumerate(field_items):
        if item.v == field_value:
            field_value_index = index
            break

    if field_index is None:
        raise ValueError(f"Field '{field_name}' not found in pivot cache.")
    if field_value_index is None:
        raise ValueError(f"Value '{field_value}' not found in pivot cache.")

    # Filter the pivot cache records
    filtered_records = []
    for record in pivot_cache.records.r:
        if record._fields[field_index].v == field_value_index:
            filtered_records.append(record)

    return filtered_records


def get_unique_values_from_pivot_worksheet(workbook, sheet_name, field) -> List[str]:
    ws = workbook[sheet_name]
    pivot = ws._pivots[0]
    cached_data = pivot.cache

    items = []

    for f in cached_data.cacheFields:
        if f.name == field:
            for item in f.sharedItems._fields:
                if hasattr(item, 'v'):
                    items.append(item.v)
                elif hasattr(item, 'n'):
                    items.append(item.n)
                elif hasattr(item, 'f'):
                    items.append(item.f)
                else:
                    items.append(item)

    if len(items) == 0:
        raise ValueError(f"Column '{field}' not found in pivot_data")

    # Use a set to get unique values
    unique_values = set(items)
    return list(unique_values)


def get_unique_values_from_pivot_cache(file_path, sheet_name, field) -> List[str]:
    """
    Returns all the unique values for a column in a pivot cache
    :param file_path: the workbook file path
    :param sheet_name: the name of the sheet
    :param field: the field to get values for
    :return: a list of unique values for the specified column in the cache
    """
    from openpyxl import load_workbook
    wb = load_workbook(file_path, data_only=True)
    return get_unique_values_from_pivot_worksheet(workbook=wb, sheet_name=sheet_name, field=field)


def get_header(file_path: str, sheet_name: str):
    header: List[str] = pd.read_excel(file_path, sheet_name=sheet_name, nrows=0).columns.tolist()
    return header


def sheet_contains_column_name(file_path: str, sheet_name: str, field: str):
    return field in get_header(file_path=file_path, sheet_name=sheet_name)


def get_unique_values_for_workbook(file_path, field) -> List[str] | None:
    """
    Gets the unique values for a field from an Excel workbook. The
    function iterates over the sheets in the workbook until it
    finds a sheet containing the data - either a regular sheet or
    a sheet containing a pivot.
    :param file_path: the path to the Excel file
    :param field: the field to get values for
    :return: a list of unique values
    :raises: ValueError if no sheets contain the field
    """
    from openpyxl import load_workbook
    wb: Workbook = load_workbook(file_path, data_only=True)
    for sheet in wb.sheetnames:
        ws = wb[sheet]
        if len(ws._pivots) > 0:
            return get_unique_values_from_pivot_worksheet(workbook=wb, sheet_name=sheet, field=field)
        else:
            if sheet_contains_column_name(file_path=file_path, sheet_name=sheet, field=field):
                return pd.read_excel(file_path, sheet_name=sheet, dtype={field: object}, usecols=[field])[field].unique().tolist()
    raise ValueError("No valid worksheet containing field values")


def get_info_sheets(workbook, field):
    info_sheets = []
    for sheet in workbook.sheetnames:
        ws = workbook[sheet]
        if len(ws._pivots) == 0:
            headers = [c.value for c in next(workbook[sheet].iter_rows(min_row=1, max_row=1))]
            if field not in headers:
                info_sheets.append(sheet)
    return info_sheets


def set_excel_active_sheet(file_path: str):
    """
    Deselects the active tabs, forcing the workbook to refresh on open.
    Useful for refreshing pivots and ensuring tabs aren't grouped on open
    :param file_path: the path to the workbook
    :return: None
    """
    from openpyxl import load_workbook

    workbook: Workbook = load_workbook(file_path)
    for sheet in workbook:
        workbook[sheet.title].views.sheetView[0].tabSelected = False
    workbook.save(file_path)
    workbook.close()


def get_worksheet_containing_field(workbook, field):
    for sheet in workbook.sheetnames:
        headers = [c.value for c in next(workbook[sheet].iter_rows(min_row=1, max_row=1))]
        if field in headers:
            return sheet


def prepend_sheet_to_workbook(source_workbook_file, target_workbook_file, sheet_name) -> None:
    """
    Prepends the specified sheet from the source workbook to the target workbook
    :param source_workbook_file: the source Excel file
    :param target_workbook_file: the target Excep file
    :param sheet_name: the sheet to prepend
    :return: None
    """
    from openpyxl import Workbook, load_workbook
    target_workbook: Workbook = load_workbook(target_workbook_file)
    source_workbook = load_workbook(source_workbook_file)
    source = source_workbook[sheet_name]
    source._parent = target_workbook
    target_workbook._add_sheet(sheet=source)
    target_workbook.move_sheet(sheet=source, offset=-(len(target_workbook.sheetnames) - 1))
    target_workbook.save(target_workbook_file)






import json
from typing import List
from mario.base import MarioBase


class Item(MarioBase):

    def __init__(self):
        super().__init__()
        self._name = ''
        self._description = ''

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        self._name = value

    @property
    def description(self):
        return self._description

    @description.setter
    def description(self, value):
        self._description = value

    def to_json(self):
        json_representation = {
            "name": self.name,
            "description": self.description
        }
        for key, value in self._properties.items():
            json_representation[key] = value
        return json_representation


class Metadata(Item):

    def __init__(self, name: str = None):
        super().__init__()
        self._items: List[Item] = []
        self.name = name
        if self.name is None:
            self.name = 'Metadata'

    def get_metadata(self, name: str):
        for item in self._items:
            if item.name == name:
                return item
        return None

    @property
    def items(self):
        return self._items

    def add_item(self, item: Item) -> None:
        self._items.append(item)

    def merge_items(self, metadata) -> None:
        for item in metadata.items:
            self.add_item(item)

    def get_hierarchies(self):
        """ Returns a list of all hierarchies in the metadata """
        hierarchies = set()
        for item in self.items:
            if item.get_property('hierarchies') is not None:
                for hierarchy in item.get_property('hierarchies'):
                    hierarchies.add(hierarchy['hierarchy'])
        return list(hierarchies)

    def get_hierarchy(self, name):
        """ Returns an ordered list of the item names in the specified hierarchy """
        items = []
        for item in self.items:
            if item.get_property('hierarchies') is not None:
                for hierarchy in item.get_property('hierarchies'):
                    if hierarchy['hierarchy'] == name:
                        items.append({'name': item.name, 'level': hierarchy['level']})

        # Sort the list of dictionaries by 'level'
        sorted_items = sorted(items, key=lambda x: x['level'])
        # Extract the 'name' values in order
        item_names_in_order = [item['name'] for item in sorted_items]
        return item_names_in_order

    def save(self, file_path: str = None) -> None:
        json_representation = {
            "collection": {
                "name": self.name,
                "items": []
            }
        }
        for prop in self.properties:
            json_representation['collection'][prop] = self.get_property(prop)

        for item in self._items:
            json_representation['collection']['items'].append(item.to_json())

        with open(file_path, mode='w', encoding='utf-8') as file:
            json.dump(json_representation, file, default=vars)


def metadata_from_json(file_path: str = None) -> Metadata:
    """ Factory method for creating a Metadata instance from a JSON file"""
    metadata = Metadata()

    with open(file_path, encoding='utf-8') as metadata_file:
        metadata_json = json.load(metadata_file)

    if 'collection' in metadata_json:
        collection = metadata_json['collection']
        metadata.name = collection['name']
        items = 'items'
        name = 'name'
    else:
        collection = metadata_json['datasource']
        metadata.name = collection['name']
        items = 'fields'
        name = 'fieldName'

    metadata.add_properties(source=collection, exclude=[name, items])

    for item in collection[items]:
        metadata_item = Item()
        metadata_item.name = item[name]
        if 'description' in item:
            metadata_item.description = item['description']
        metadata_item.add_properties(source=item, exclude=[name, 'description'])
        metadata.add_item(metadata_item)

    return metadata


def metadata_from_manifest(file_path=None) -> Metadata:
    """ Factory method for creating a Metadata instance from a JSON file in manifest format"""
    metadata = Metadata()

    with open(file_path, encoding='utf-8') as metadata_file:
        metadata_json = json.load(metadata_file)

    collection = metadata_json
    metadata.name = metadata_json['datasource']
    items = 'items'
    name = 'fieldName'

    metadata.add_properties(source=collection, exclude=[name, items])

    for item in collection[items]:
        metadata_item = Item()
        metadata_item.name = item[name]
        if 'description' in item:
            metadata_item.description = item['description']
        metadata_item.add_properties(source=item, exclude=[name, 'description'])
        metadata.add_item(metadata_item)

    # For TDSA manifests, need to also add the measure as an item
    if 'measure' in metadata_json:
        measure_metadata_item = Item()
        measure_metadata_item.name = metadata_json['measure']
        metadata.add_item(measure_metadata_item)

    return metadata


def metadata_from_excel(
        file_path: str = None,
        name: str = "Metdata",
        sheet_name: str = 'Pick list',
        field_name_column: str = 'Field + Options combined'
) -> Metadata:
    """ Factory method for creating a Metadata instance from an Excel file"""
    import pandas as pd
    import re
    pick_list = pd.read_excel(open(file_path, 'rb'), sheet_name=sheet_name, skiprows=1, header=0)
    pick_list.reset_index()
    pick_list.fillna('', inplace=True)
    pick_list.dropna(how='all', axis=1, inplace=True)

    metadata = Metadata()
    metadata.name = name
    metadata.set_property('source', file_path)
    for index, row in pick_list.iterrows():
        if isinstance(row[field_name_column], str):
            data_item = Item()
            data_item.name = row[field_name_column]
            for key, value in row.to_dict().items():
                data_item.set_property(key, value)
            # Domain splits
            pattern = re.compile("\\(([^)]*)\\)")
            found = re.findall(pattern, data_item.name)
            if found:
                domain = [x.strip() for x in found[0].split('/')]
                if len(domain) > 1:
                    data_item.set_property('domain', domain)
            metadata.add_item(data_item)

    return metadata



from mario.query_builder import ViewBasedQueryBuilder, SubsetQueryBuilder, get_formatted_query
from mario.data_extractor import Configuration
from mario.metadata import Metadata, Item
from mario.dataset_specification import DatasetSpecification, Constraint


def test_query_builder():
    dataset_specification = DatasetSpecification()

    metadata = Metadata()

    configuration = Configuration()
    configuration.schema = 'dbo'
    configuration.view = 'v_extract_test'

    query_builder = ViewBasedQueryBuilder(
        dataset_specification=dataset_specification,
        metadata=metadata,
        configuration=configuration
    )

    query = query_builder.create_query()
    assert query[0] == 'SELECT * FROM "dbo"."v_extract_test"'


def test_subset_query_builder():
    dataset_specification = DatasetSpecification()
    dataset_specification.dimensions.append('fruit')
    dataset_specification.measures.append('volume')

    metadata = Metadata()
    item = Item()
    item.name = 'fruit'
    metadata.add_item(item)
    item = Item()
    item.name = 'volume'
    metadata.add_item(item)

    configuration = Configuration()
    configuration.schema = 'dbo'
    configuration.view = 'v_extract_test'

    query_builder = SubsetQueryBuilder(
        dataset_specification=dataset_specification,
        metadata=metadata,
        configuration=configuration
    )

    query = query_builder.create_query()
    assert query[0] == 'SELECT "fruit",SUM("volume") "volume" FROM "dbo"."v_extract_test" ' \
                       'GROUP BY "fruit"'


def test_subset_query_builder_multiple_measures():
    dataset_specification = DatasetSpecification()
    dataset_specification.dimensions.append('fruit')
    dataset_specification.measures.append('volume')
    dataset_specification.measures.append('weight')

    metadata = Metadata()
    for field in ['fruit', 'volume', 'weight']:
        item = Item()
        item.name = field
        metadata.add_item(item)

    configuration = Configuration()
    configuration.schema = 'dbo'
    configuration.view = 'v_extract_test'

    query_builder = SubsetQueryBuilder(
        dataset_specification=dataset_specification,
        metadata=metadata,
        configuration=configuration
    )

    query = query_builder.create_query()
    assert query[0] == 'SELECT "fruit",SUM("volume") "volume",SUM("weight") "weight" ' \
                       'FROM "dbo"."v_extract_test" GROUP BY "fruit"'


def test_subset_query_builder_with_constraint():
    dataset_specification = DatasetSpecification()
    dataset_specification.dimensions.append('fruit')
    dataset_specification.dimensions.append('colour')
    dataset_specification.measures.append('volume')
    constraint = Constraint()
    constraint.item = 'colour'
    constraint.allowed_values=['yellow', 'orange']
    dataset_specification.constraints.append(constraint)

    metadata = Metadata()
    for field in ['fruit', 'colour', 'volume']:
        item = Item()
        item.name = field
        metadata.add_item(item)

    configuration = Configuration()
    configuration.schema = 'dbo'
    configuration.view = 'v_extract_test'

    query_builder = SubsetQueryBuilder(
        dataset_specification=dataset_specification,
        metadata=metadata,
        configuration=configuration
    )

    query = query_builder.create_query()
    assert query[0] == 'SELECT "fruit","colour",SUM("volume") "volume" FROM "dbo"."v_extract_test" ' \
                       'WHERE "colour" IN (%(colour0)s,%(colour1)s) GROUP BY "fruit","colour"'
    assert query[1] == {'colour0': 'yellow', 'colour1': 'orange'}


def test_subset_query_builder_with_calculated_field():
    dataset_specification = DatasetSpecification()
    dataset_specification.dimensions.append('fruit')
    dataset_specification.measures.append('price')
    dataset_specification.measures.append('profit')
    dataset_specification.measures.append('profit ratio')

    metadata = Metadata()
    for field in ['fruit', 'price', 'profit', 'profit ratio']:
        item = Item()
        item.name = field
        metadata.add_item(item)

    metadata.get_metadata('profit ratio').set_property('formula', 'profit/price')

    configuration = Configuration()
    configuration.schema = 'dbo'
    configuration.view = 'v_extract_test'

    query_builder = SubsetQueryBuilder(
        dataset_specification=dataset_specification,
        metadata=metadata,
        configuration=configuration
    )

    query = query_builder.create_query()
    assert query[0] == 'SELECT "fruit",SUM("price") "price",SUM("profit") "profit" ' \
                       'FROM "dbo"."v_extract_test" GROUP BY "fruit"'


def test_formatted_subset_query_builder_with_constraint():
    dataset_specification = DatasetSpecification()
    dataset_specification.dimensions.append('fruit')
    dataset_specification.dimensions.append('colour')
    dataset_specification.measures.append('volume')
    constraint = Constraint()
    constraint.item = 'colour'
    constraint.allowed_values=['yellow', 'orange']
    dataset_specification.constraints.append(constraint)

    metadata = Metadata()
    for field in ['fruit', 'colour', 'volume']:
        item = Item()
        item.name = field
        metadata.add_item(item)

    configuration = Configuration()
    configuration.schema = 'dbo'
    configuration.view = 'v_extract_test'

    query_builder = SubsetQueryBuilder(
        dataset_specification=dataset_specification,
        metadata=metadata,
        configuration=configuration
    )

    query = query_builder.create_query()
    formatted_query = get_formatted_query(query[0], query[1])
    assert query[0] == 'SELECT "fruit","colour",SUM("volume") "volume" FROM "dbo"."v_extract_test" ' \
                       'WHERE "colour" IN (%(colour0)s,%(colour1)s) GROUP BY "fruit","colour"'
    assert query[1] == {'colour0': 'yellow', 'colour1': 'orange'}
    assert formatted_query == 'SELECT "fruit","colour",SUM("volume") "volume" FROM "dbo"."v_extract_test" ' \
                              'WHERE "colour" IN (\'yellow\',\'orange\') GROUP BY "fruit","colour"'



"""
Hyper utility functions
"""
import logging
from typing import List

log = logging.getLogger(__name__)


def get_default_table_and_schema(hyper_path) -> (str, str):
    """
    Get the default table and schema from a hyper, assuming its
    a single-table hyper
    :param hyper_path:
    :return:
    """
    from tableauhyperapi import HyperProcess, Connection, Telemetry
    tables = []
    with HyperProcess(Telemetry.DO_NOT_SEND_USAGE_DATA_TO_TABLEAU, 'test') as hyper:
        with Connection(hyper.endpoint, hyper_path) as connection:
            # The `connection.catalog` provides us with access to the meta-data we are interested in
            catalog = connection.catalog
            # Iterate over all schemas
            schemas = catalog.get_schema_names()
            for schema_name in schemas:
                # For each schema, iterate over all tables
                schema_tables = catalog.get_table_names(schema=schema_name)
                if len(schema_tables) > 0:
                    tables = schema_tables
    table = tables[0].name.unescaped
    schema = tables[0].schema_name
    if schema is None:
        schema = 'public'
    else:
        schema = schema.name.unescaped
    return schema, table


def check_column_exists(hyper_file_path: str, column: str, schema='public', table='default'):
    columns = get_column_list(hyper_file_path=hyper_file_path, schema=schema, table=table)
    return column in columns


def get_column_list(hyper_file_path: str, schema='public', table='default'):
    """
    Get the list of columns in a table
    :param hyper_file_path:
    :param schema:
    :param table:
    :return:
    """
    from tableauhyperapi import HyperProcess, Telemetry, Connection, TableName
    with HyperProcess(telemetry=Telemetry.SEND_USAGE_DATA_TO_TABLEAU) as hyper:
        with Connection(endpoint=hyper.endpoint, database=hyper_file_path) as connection:
            table_name = TableName(schema, table)
            table_definition = connection.catalog.get_table_definition(table_name)
            columns = [col.name.unescaped for col in table_definition.columns]
    return columns


def filter_column_list(hyper_file_path: str, columns:List[str], schema='public', table='default'):
    """
    Filters a list of columns, removing any that aren't present in the hyper
    :param hyper_file_path:
    :param columns:
    :return:
    """
    filtered_columns = []
    td_columns = get_column_list(hyper_file_path=hyper_file_path, schema=schema, table=table)
    for column in columns:
        for td_column in td_columns:
            if td_column == column:
                filtered_columns.append(column)
    return filtered_columns


def get_table(hyper_path: str, schema='public', table='default'):
    from tableauhyperapi import HyperProcess, Telemetry, Connection, TableName, TableDefinition
    with HyperProcess(Telemetry.DO_NOT_SEND_USAGE_DATA_TO_TABLEAU, 'test') as hyper:
        with Connection(hyper.endpoint, hyper_path) as connection:
            table_name_tuple = TableName(schema, table)
            table: TableDefinition = connection.catalog.get_table_definition(table_name_tuple)
    return table


def get_total(hyper_file_path: str, schema='public', table='default', measure='FPE'):
    from tableauhyperapi import HyperProcess, Telemetry, Connection
    with HyperProcess(telemetry=Telemetry.SEND_USAGE_DATA_TO_TABLEAU) as hyper:
        with Connection(endpoint=hyper.endpoint, database=hyper_file_path) as connection:
            total = connection.execute_scalar_query(f"SELECT SUM(\"{measure}\") FROM \"{schema}\".\"{table}\"")
    return total


def get_row_count(hyper_file_path: str, schema='public', table='default'):
    from tableauhyperapi import HyperProcess, Telemetry, Connection
    with HyperProcess(telemetry=Telemetry.SEND_USAGE_DATA_TO_TABLEAU) as hyper:
        with Connection(endpoint=hyper.endpoint, database=hyper_file_path) as connection:
            row_count = connection.execute_scalar_query(f"SELECT COUNT(*) FROM \"{schema}\".\"{table}\"")
    return row_count


def add_row_numbers_to_hyper(input_hyper_file_path: str, schema='public', table='default'):
    from tableauhyperapi import HyperProcess, Telemetry, Connection, TableName, TableDefinition, SqlType

    drop_columns_from_hyper(hyper_file_path=input_hyper_file_path, columns_to_drop=['row_number'])

    with HyperProcess(telemetry=Telemetry.SEND_USAGE_DATA_TO_TABLEAU) as hyper:
        with Connection(endpoint=hyper.endpoint, database=input_hyper_file_path) as connection:
            input_table_name = TableName(schema, table)
            output_table_name = TableName(schema, table + '_new')

            output_table_definition = TableDefinition(output_table_name,
                                                      connection.catalog.get_table_definition(
                                                          input_table_name).columns)
            output_table_definition.add_column(TableDefinition.Column('row_number', SqlType.big_int()))
            connection.catalog.create_table(output_table_definition)
            log.debug("Created output table")

        with Connection(endpoint=hyper.endpoint, database=input_hyper_file_path) as connection:
            query = f"""
                    WITH CTE AS (
                        SELECT *, ROW_NUMBER() OVER() AS row_number
                        FROM {input_table_name.name}
                    )
                    INSERT INTO {output_table_name.name}
                        SELECT * FROM CTE
                    """
            connection.execute_command(query)
            log.debug(f"Added row numbers to {input_table_name} and output to {output_table_name}")

        with Connection(endpoint=hyper.endpoint, database=input_hyper_file_path) as connection:
            drop_table_query = f"""
                    DROP TABLE {input_table_name.name}
            """
            connection.execute_command(drop_table_query)
            log.debug("Dropped old table")
            rename_table_query = f"""
                    ALTER TABLE {output_table_name.name} RENAME TO {input_table_name.name}
            """
            connection.execute_command(rename_table_query)
            log.debug("Renamed new table")


def filter_hyper_by_row_number(input_hyper_file_path: str, output_hyper_file_path: str, row_numbers, schema='public', table='default'):
    where = f"row_number NOT IN ({','.join([str(x) for x in row_numbers])})"
    filter_hyper(input_hyper_file_path=input_hyper_file_path,
                 output_hyper_file_path=output_hyper_file_path,
                 where=where,
                 schema=schema,
                 table=table)


def filter_hyper(input_hyper_file_path: str, output_hyper_file_path: str, where: str = "1=1", schema='public', table='default'):
    from tableauhyperapi import HyperProcess, Telemetry, Connection, TableName, TableDefinition
    with HyperProcess(telemetry=Telemetry.SEND_USAGE_DATA_TO_TABLEAU) as hyper:
        with Connection(endpoint=hyper.endpoint) as connection:
            # Connect to the input and output databases
            # Create the output Hyper file or overwrite it
            catalog = connection.catalog
            catalog.drop_database_if_exists(output_hyper_file_path)
            catalog.create_database(output_hyper_file_path)
            catalog.attach_database(output_hyper_file_path, alias="output_database")
            catalog.attach_database(input_hyper_file_path, alias="input_database")

            input_table_name = TableName("input_database", schema, table)
            output_table_name = TableName("output_database", schema, table)
            output_table_definition = TableDefinition(output_table_name, catalog.get_table_definition(
                input_table_name).columns)
            catalog.create_schema_if_not_exists(output_table_name.schema_name)
            catalog.create_table(output_table_definition)
            connection.execute_command(
                f"INSERT INTO {output_table_name} "
                f"(SELECT * FROM {input_table_name} "
                f"WHERE {where} "
                f")")
    log.debug(f"Filtered {input_hyper_file_path} into {output_hyper_file_path}")


def defragment_hyper(input_hyper_file_path: str, output_hyper_file_path: str, schema='public', table='default'):
    filter_hyper(
        input_hyper_file_path=input_hyper_file_path,
        output_hyper_file_path=output_hyper_file_path,
        schema=schema,
        table=table
    )


def drop_columns_from_hyper(hyper_file_path, columns_to_drop=None, columns_to_keep=None, schema='public', table='default'):
    """
    Because there is no way to drop a column in a hyper, we have this stupid workaround
    """
    from tableauhyperapi import HyperProcess, Telemetry, Connection, TableName, TableDefinition, escape_name

    if columns_to_drop is not None:
        columns_to_drop = filter_column_list(
            hyper_file_path=hyper_file_path,
            columns=columns_to_drop,
            schema=schema,
            table=table
        )
    elif columns_to_keep is not None:
        all_columns = get_column_list(hyper_file_path=hyper_file_path, schema=schema, table=table)
        columns_to_drop = []
        for column in all_columns:
            if column not in columns_to_keep:
                columns_to_drop.append(column)
    else:
        raise ValueError("Must supply a list of columns to keep or to drop")

    if columns_to_drop is None or len(columns_to_drop) == 0:
        log.debug(f"No columns to drop from dataset.")
    else:
        log.info(f"- Dropping {len(columns_to_drop)} columns from dataset.")
        with HyperProcess(telemetry=Telemetry.SEND_USAGE_DATA_TO_TABLEAU) as hyper:
            with Connection(endpoint=hyper.endpoint, database=hyper_file_path) as connection:
                # Get the existing table definition
                existing_table_definition = connection.catalog.get_table_definition(TableName(schema, table))

                # Define the new table schema without the columns to drop
                new_columns = [
                    col for col in existing_table_definition.columns if col.name.unescaped not in columns_to_drop
                ]
                new_table_definition = TableDefinition(
                    table_name=TableName(schema, f"{table}_new"),
                    columns=new_columns
                )
                if len(new_columns) != len(existing_table_definition.columns) - len(columns_to_drop):
                    raise ValueError(f"One or more columns specified to drop ({columns_to_drop})"
                                     f" are not present in the dataset.")

            # Create the new table
            with Connection(endpoint=hyper.endpoint, database=hyper_file_path) as connection:
                connection.catalog.create_table(new_table_definition)

            # Copy data
            with Connection(endpoint=hyper.endpoint, database=hyper_file_path) as connection:
                select_columns = ", ".join([escape_name(col.name.unescaped) for col in new_columns])
                query = f"INSERT INTO \"{schema}\".\"{table}_new\" (SELECT {select_columns} FROM \"{schema}\".\"{table}\")"
                connection.execute_command(query)

            with Connection(endpoint=hyper.endpoint, database=hyper_file_path) as connection:
                # Drop the old table
                connection.execute_command(f"DROP TABLE \"{schema}\".\"{table}\"")

                # Rename the new table to the old table name
                connection.execute_command(
                    f"ALTER TABLE \"{schema}\".\"{table}_new\" RENAME TO \"{table}\"")


def create_random_sample(hyper_file_path, sample_size, schema='public', table='default'):
    from tableauhyperapi import HyperProcess, Connection, Telemetry

    with HyperProcess(telemetry=Telemetry.SEND_USAGE_DATA_TO_TABLEAU) as hyper:
        with Connection(endpoint=hyper.endpoint, database=hyper_file_path) as connection:

            # Create a temporary table with the random sample
            connection.execute_command(f"""
                CREATE TABLE random_sample AS
                SELECT *
                FROM \"{schema}\".\"{table}\"
                ORDER BY RANDOM()
                LIMIT {sample_size}
            """)

            connection.execute_command(f"DROP TABLE \"{schema}\".\"{table}\"")
            connection.execute_command(f"ALTER TABLE random_sample RENAME TO \"{table}\"")


def concatenate_hypers(hyper_file_path1, hyper_file_path2, output_hyper_file_path, schema='public', table='default'):
    from pantab import frame_from_hyper, frame_to_hyper
    from tableauhyperapi import TableName
    import pandas as pd

    df1 = frame_from_hyper(hyper_file_path1, table=TableName(schema, table))
    df2 = frame_from_hyper(hyper_file_path2, table=TableName(schema, table))

    df_merged = pd.concat([df1, df2], axis=1)
    frame_to_hyper(df=df_merged, database=output_hyper_file_path, table=TableName(schema, table))